"""Unit tests for reviewboard.reviews.manager.ReviewGroupManager."""

from __future__ import annotations

from itertools import chain
from typing import Dict, List, Optional, Sequence, TYPE_CHECKING, Tuple, Union

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, User
from djblets.testing.decorators import add_fixtures

from reviewboard.reviews.models import Group
from reviewboard.reviews.testing.queries.review_groups import (
    get_review_groups_accessible_equeries,
    get_review_groups_accessible_ids_equeries)
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from typelets.funcs import KwargsDict

    from reviewboard.site.models import AnyOrAllLocalSites

    _MixinParent = TestCase
else:
    _MixinParent = object


class AccessibleTestsMixin(_MixinParent):
    """Mixins for review group accessibility unit tests.

    Version Added:
        5.0.7
    """

    def _create_accessible_review_group_data(
        self,
        *,
        user: Union[AnonymousUser, User],
        with_local_sites: bool = False,
        with_member: bool = False,
        group_kwargs: KwargsDict = {},
    ) -> Tuple[Sequence[Group], Optional[LocalSite]]:
        """Create test review group data for accessibility checks.

        This will create review groups on the global site and, optionally,
        two Local Sites. The provided user may optionally be granted membership
        directly.

        Args:
            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to associate with any membership lists.

            with_local_sites (bool, optional):
                Whether to create Local Sites.

            with_member (bool, optional):
                Whether to make the user a member of the review group.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (list of reviewboard.reviews.models.group.Group):
                    The list of created review groups, in order.

                1 (reviewboard.site.models.LocalSite):
                    The first Local Site created, or ``None`` if not creating
                    Local Sites.
        """
        groups_by_site: Dict[Optional[LocalSite], List[Group]] = {}
        local_sites: List[Optional[LocalSite]] = []
        local_site_kwargs: KwargsDict = {}
        group_i = 1

        if user.is_authenticated:
            # Satisfy the type checker.
            assert isinstance(user, User)

            local_site_kwargs['users'] = [user]

        if with_local_sites:
            # Create the Local Sites
            local_sites += [
                self.create_local_site(name='test-site-1',
                                       **local_site_kwargs),
                self.create_local_site(name='test-site-2',
                                       **local_site_kwargs),
            ]

        # Add the global site last.
        local_sites.append(None)

        for local_site in local_sites:
            groups_by_site[local_site] = [
                self.create_review_group(local_site=local_site,
                                         name=f'repo{group_i:02}',
                                         invite_only=False,
                                         visible=True),
                self.create_review_group(local_site=local_site,
                                         name=f'repo{group_i + 1:02}',
                                         invite_only=False,
                                         visible=False),
                self.create_review_group(local_site=local_site,
                                         name=f'repo{group_i + 2:02}',
                                         invite_only=True,
                                         visible=True),
                self.create_review_group(local_site=local_site,
                                         name=f'repo{group_i + 3:02}',
                                         invite_only=True,
                                         visible=False),
            ]
            group_i += 4

        if with_member and user.is_authenticated:
            # Satisfy the type checker.
            assert isinstance(user, User)

            for local_site in local_sites:
                user.review_groups.add(*groups_by_site[local_site])

        return (
            list(chain.from_iterable(groups_by_site.values())),
            local_sites[0],
        )


class AccessibleTests(AccessibleTestsMixin, TestCase):
    """Unit tests for ReviewGroupManager.accessible()."""

    def test_with_anonymous(self) -> None:
        """Testing Group.objects.accessible with anonymous and
        visible_only=False
        """
        user = AnonymousUser()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible(
            groups=[
                groups[0],
                groups[1],
            ],
            user=user,
            visible_only=False)

    def test_with_anonymous_visible_only(self) -> None:
        """Testing Group.objects.accessible with anonymous and
        visible_only=True
        """
        user = AnonymousUser()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible(
            groups=[groups[0]],
            user=user,
            visible_only=True)

    #
    # Superuser tests
    #

    def test_with_superuser(self) -> None:
        """Testing Group.objects.accessible with superuser and
        visible_only=False
        """
        user = self.create_user(is_superuser=True)
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible(
            groups=groups,
            user=user,
            visible_only=False)

    def test_with_superuser_visible_only(self) -> None:
        """Testing Group.objects.accessible with superuser and
        visible_only=True
        """
        user = self.create_user(is_superuser=True)
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible(
            groups=[
                groups[0],
                groups[2],
            ],
            user=user,
            visible_only=True)

    def test_with_superuser_local_site(self) -> None:
        """Testing Group.objects.accessible with superuser, Local Site,
        and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=groups[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_superuser_local_site_visible_only(self) -> None:
        """Testing Group.objects.accessible with superuser, Local Site,
        and visible_only=True
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=[
                groups[0],
                groups[2],
            ],
            user=user,
            local_site=local_site,
            visible_only=True)

    def test_with_superuser_local_site_all(self) -> None:
        """Testing Group.objects.accessible with superuser, all Local
        Sites, and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=groups,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_superuser_local_site_all_visible_only(self) -> None:
        """Testing Group.objects.accessible with superuser, all Local
        Sites, and visible_only=True
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=[
                groups[0],
                groups[2],
                groups[4],
                groups[6],
                groups[8],
                groups[10],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Non-member tests
    #

    def test_with_non_member(self) -> None:
        """Testing Group.objects.accessible with non-member and
        visible_only=False
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible(
            groups=[
                groups[0],
                groups[1],
            ],
            user=user,
            visible_only=False)

    def test_with_non_member_visible_only(self) -> None:
        """Testing Group.objects.accessible with non-member and
        visible_only=True
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible(
            groups=[groups[0]],
            user=user,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_non_member_local_site(self) -> None:
        """Testing Group.objects.accessible with non-member, Local Site,
        and visible_only=False
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=[
                groups[0],
                groups[1],
            ],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_non_member_local_site_visible_only(self) -> None:
        """Testing Group.objects.accessible with non-member, Local Site,
        and visible_only=True
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=[groups[0]],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_non_member_local_site_all(self) -> None:
        """Testing Group.objects.accessible with non-member, all Local
        Sites, and visible_only=False
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=[
                groups[0],
                groups[1],
                groups[4],
                groups[5],
                groups[8],
                groups[9],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_non_member_local_site_all_visible_only(self) -> None:
        """Testing Group.objects.accessible with non-member, all Local
        Sites, and visible_only=True
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            groups=[
                groups[0],
                groups[4],
                groups[8],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member tests
    #

    def test_with_member(self) -> None:
        """Testing Group.objects.accessible with member and
        visible_only=False
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_member=True)[0]

        self._test_accessible(
            groups=groups[:4],
            user=user,
            visible_only=False)

    def test_with_member_visible_only(self) -> None:
        """Testing Group.objects.accessible with member and
        visible_only=True
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_member=True)[0]

        # Note that members can see hidden review groups. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            groups=groups,
            user=user,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_member_local_site(self) -> None:
        """Testing Group.objects.accessible with member, Local Site,
        and visible_only=False
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        self._test_accessible(
            groups=groups[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_member_local_site_visible_only(self) -> None:
        """Testing Group.objects.accessible with member, Local Site,
        and visible_only=True
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        # Note that members can see hidden review groups. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            groups=groups[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_member_local_site_all(self) -> None:
        """Testing Group.objects.accessible with member, all Local
        Sites, and visible_only=False
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True,
            with_member=True)[0]

        self._test_accessible(
            groups=groups,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_member_local_site_all_visible_only(self) -> None:
        """Testing Group.objects.accessible with member, all Local
        Sites, and visible_only=True
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)[0]

        self._test_accessible(
            groups=[
                groups[0],
                groups[4],
                groups[8],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_show_all_local_sites(self) -> None:
        """Testing Group.objects.accessible with show_all_local_sites=True
        """
        user = self.create_user()

        group = self.create_review_group(with_local_site=True)
        group.local_site.users.add(user)

        self.assertIn(
            group,
            Group.objects.accessible(user, local_site=group.local_site))
        self.assertIn(
            group,
            Group.objects.accessible(user, local_site=LocalSite.ALL))

    def _test_accessible(
        self,
        *,
        groups: Sequence[Group],
        user: Union[AnonymousUser, User],
        visible_only: bool,
        local_site: AnyOrAllLocalSites = None,
        has_view_invite_only_groups_perm: bool = False,
    ) -> None:
        """Test the results of accessible(), using the given settings.

        Version Added:
            5.0.7

        Args:
            groups (list of reviewboard.revies.models.group.Group):
                The review groups to expect in results.

            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to check for access.

            visible_only (bool):
                Whether to check for visible groups only.

            local_site (reviewboard.site.models.LocalSite, optional):
                The optional Local Site to check.

            expect_in (bool, optional):
                Whether to expect the provided group's ID to be in the
                accessible IDs list.

            has_view_invite_only_groups_perm (bool, optional):
                Whether to expect the
                ``reviews.has_view_invite_only_groups_perm`` permission to
                be set.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        # Prime the caches.
        if user.is_authenticated:
            user.get_local_site_stats()
            user.get_profile()

            if local_site is not LocalSite.ALL:
                user.get_site_profile(local_site=local_site)

        equeries = get_review_groups_accessible_equeries(
            user=user,
            visible_only=visible_only,
            has_view_invite_only_groups_perm=has_view_invite_only_groups_perm,
            local_site=local_site)

        with assert_queries(equeries):
            accessible = list(Group.objects.accessible(
                user,
                visible_only=visible_only,
                local_site=local_site))

        self.assertEqual(accessible, groups)


class AccessibleIDsTests(AccessibleTestsMixin, TestCase):
    """Unit tests for ReviewGroupManager.accessible_ids()."""

    #
    # Anonymous tests
    #

    def test_with_anonymous(self) -> None:
        """Testing Group.objects.accessible_ids with anonymous and
        visible_only=False
        """
        user = AnonymousUser()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[1],
            ],
            user=user,
            visible_only=False)

    def test_with_anonymous_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with anonymous and
        visible_only=True
        """
        user = AnonymousUser()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible_ids(
            groups=[groups[0]],
            user=user,
            visible_only=True)

    #
    # Superuser tests
    #

    def test_with_superuser(self) -> None:
        """Testing Group.objects.accessible_ids with superuser and
        visible_only=False
        """
        user = self.create_user(is_superuser=True)
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible_ids(
            groups=groups,
            user=user,
            visible_only=False)

    def test_with_superuser_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with superuser and
        visible_only=True
        """
        user = self.create_user(is_superuser=True)
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[2],
            ],
            user=user,
            visible_only=True)

    def test_with_superuser_local_site(self) -> None:
        """Testing Group.objects.accessible_ids with superuser, Local Site,
        and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=groups[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_superuser_local_site_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with superuser, Local Site,
        and visible_only=True
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[2],
            ],
            user=user,
            local_site=local_site,
            visible_only=True)

    def test_with_superuser_local_site_all(self) -> None:
        """Testing Group.objects.accessible_ids with superuser, all Local
        Sites, and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=groups,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_superuser_local_site_all_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with superuser, all Local
        Sites, and visible_only=True
        """
        user = self.create_user(is_superuser=True)
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[2],
                groups[4],
                groups[6],
                groups[8],
                groups[10],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Non-member tests
    #

    def test_with_non_member(self) -> None:
        """Testing Group.objects.accessible_ids with non-member and
        visible_only=False
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[1],
            ],
            user=user,
            visible_only=False)

    def test_with_non_member_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with non-member and
        visible_only=True
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(user=user)[0]

        self._test_accessible_ids(
            groups=[groups[0]],
            user=user,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_non_member_local_site(self) -> None:
        """Testing Group.objects.accessible_ids with non-member, Local Site,
        and visible_only=False
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[1],
            ],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_non_member_local_site_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with non-member, Local Site,
        and visible_only=True
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=[groups[0]],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_non_member_local_site_all(self) -> None:
        """Testing Group.objects.accessible_ids with non-member, all Local
        Sites, and visible_only=False
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[1],
                groups[4],
                groups[5],
                groups[8],
                groups[9],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_non_member_local_site_all_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with non-member, all Local
        Sites, and visible_only=True
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[4],
                groups[8],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member tests
    #

    def test_with_member(self) -> None:
        """Testing Group.objects.accessible_ids with member and
        visible_only=False
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_member=True)[0]

        self._test_accessible_ids(
            groups=groups[:4],
            user=user,
            visible_only=False)

    def test_with_member_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with member and
        visible_only=True
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_member=True)[0]

        # Note that members can see hidden review groups. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            groups=groups,
            user=user,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_member_local_site(self) -> None:
        """Testing Group.objects.accessible_ids with member, Local Site,
        and visible_only=False
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        self._test_accessible_ids(
            groups=groups[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_member_local_site_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with member, Local Site,
        and visible_only=True
        """
        user = self.create_user()
        groups, local_site = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        # Note that members can see hidden review groups. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            groups=groups[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_member_local_site_all(self) -> None:
        """Testing Group.objects.accessible_ids with member, all Local
        Sites, and visible_only=False
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True,
            with_member=True)[0]

        self._test_accessible_ids(
            groups=groups,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_member_local_site_all_visible_only(self) -> None:
        """Testing Group.objects.accessible_ids with member, all Local
        Sites, and visible_only=True
        """
        user = self.create_user()
        groups = self._create_accessible_review_group_data(
            user=user,
            with_local_sites=True)[0]

        self._test_accessible_ids(
            groups=[
                groups[0],
                groups[4],
                groups[8],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_show_all_local_sites(self) -> None:
        """Testing Group.objects.accessible_ids with show_all_local_sites=True
        """
        user = self.create_user()

        group = self.create_review_group(with_local_site=True)
        group.local_site.users.add(user)

        self.assertIn(
            group.pk,
            Group.objects.accessible_ids(user, local_site=group.local_site))
        self.assertIn(
            group.pk,
            Group.objects.accessible_ids(user, local_site=LocalSite.ALL))

    def _test_accessible_ids(
        self,
        *,
        groups: Sequence[Group],
        user: Union[AnonymousUser, User],
        visible_only: bool,
        local_site: AnyOrAllLocalSites = None,
        has_view_invite_only_groups_perm: bool = False,
    ) -> None:
        """Test the results of accessible_ids(), using the given settings.

        Version Added:
            5.0.7

        Args:
            groups (list of reviewboard.revies.models.group.Group):
                The review groups to expect in results.

            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to check for access.

            visible_only (bool):
                Whether to check for visible groups only.

            local_site (reviewboard.site.models.LocalSite, optional):
                The optional Local Site to check.

            expect_in (bool, optional):
                Whether to expect the provided group's ID to be in the
                accessible IDs list.

            has_view_invite_only_groups_perm (bool, optional):
                Whether to expect the
                ``reviews.has_view_invite_only_groups_perm`` permission to
                be set.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        # Prime the caches.
        if user.is_authenticated:
            user.get_local_site_stats()
            user.get_profile()

            if local_site is not LocalSite.ALL:
                user.get_site_profile(local_site=local_site)

        equeries = get_review_groups_accessible_ids_equeries(
            user=user,
            visible_only=visible_only,
            has_view_invite_only_groups_perm=has_view_invite_only_groups_perm,
            local_site=local_site)

        with assert_queries(equeries):
            accessible_ids = Group.objects.accessible_ids(
                user,
                visible_only=visible_only,
                local_site=local_site)

        self.assertEqual(
            accessible_ids,
            [
                group.pk
                for group in groups
            ])



"""Base support for action unit tests.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from reviewboard.actions import (ActionPlacement,
                                 AttachmentPoint,
                                 BaseAction,
                                 BaseGroupAction,
                                 BaseMenuAction)
from reviewboard.actions.registry import (ActionAttachmentPointsRegistry,
                                          ActionsRegistry)
from reviewboard.actions.renderers import ButtonActionRenderer

if TYPE_CHECKING:
    from collections.abc import Iterator

    from django.template import Context
    from typelets.django.json import SerializableDjangoJSONDict

    from reviewboard.actions.base import ActionAttachmentPoint


class TestAction(BaseAction):
    """Basic action for testing.

    Version Added:
        8.0
    """

    action_id = 'test'
    label = 'Test Action 1'


class TestHeaderAction(BaseAction):
    """Basic header action for testing.

    Version Added:
        8.0
    """

    action_id = 'header-action'
    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER),
    ]


class TestGroupAction(BaseGroupAction):
    """Basic group action for testing.

    Version Added:
        8.0
    """

    action_id = 'group-action'
    label = 'Test Group'

    children = [
        'group-item-2-action',
        'group-item-1-action',
    ]


class TestGroupActionWithSubgroups(BaseGroupAction):
    """Group action with subgroups for testing.

    Version Added:
        8.0
    """

    action_id = 'group-with-subgroups-action'
    label = 'Test Group with Subgroups'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER),
    ]


class TestSubgroupAction(BaseGroupAction):
    """Subgroup action for testing.

    Version Added:
        8.0
    """

    action_id = 'subgroup-action'
    label = 'Subgroup'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=TestGroupActionWithSubgroups.action_id),
    ]


class TestGroupItemAction1(BaseAction):
    """Basic group item action for testing.

    Version Added:
        8.0
    """

    action_id = 'group-item-1-action'
    label = 'Group Item 1'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=TestGroupAction.action_id),
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=TestGroupActionWithSubgroups.action_id),
    ]


class TestGroupItemAction2(BaseAction):
    """Basic group item action for testing.

    Version Added:
        8.0
    """

    action_id = 'group-item-2-action'
    label = 'Group Item 2'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=TestGroupAction.action_id),
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=TestSubgroupAction.action_id),
    ]


class TestGroupItemAction3(BaseAction):
    """Basic group item action for testing.

    Version Added:
        8.0
    """

    action_id = 'group-item-3-action'
    label = 'Group Item 3'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id=TestGroupAction.action_id),
    ]


class TestMenuAction(BaseMenuAction):
    """Basic menu action for testing.

    Version Added:
        8.0
    """

    action_id = 'menu-action'
    label = 'Test Menu'


class TestMenuItemAction(BaseAction):
    """Basic menu item action for testing.

    Version Added:
        8.0
    """

    action_id = 'menu-item-action'
    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id='menu-action'),
    ]

    label = 'Menu Item Action 1'
    icon_class = 'my-icon'
    verbose_label = 'Verbose Menu Item Action 1'
    description = ['Menu Item 1 description.']


class TestNestedMenuAction(BaseMenuAction):
    """Basic nested menu action for testing.

    Version Added:
        8.0
    """

    action_id = 'nested-menu-action'
    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id='menu-action'),
    ]


class TestNested2MenuAction(BaseMenuAction):
    """Second-level nested menu action for testing.

    Version Added:
        8.0
    """

    action_id = 'nested-2-menu-action'
    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id='nested-menu-action'),
    ]


class TooDeeplyNestedAction(BaseAction):
    """Third-level (too-deep) nested menu action for testing.

    Version Added:
        8.0
    """

    action_id = 'nested-3-action'
    placements = [
        ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                        parent_id='nested-2-menu-action'),
    ]


class SpecialButtonActionRenderer(ButtonActionRenderer):
    """Action renderer with additional button state for testing purposes.

    Version Added:
        8.0
    """

    js_view_class = 'SpecialButtonActionView'

    def get_js_view_data(
        self,
        *,
        context: Context,
    ) -> SerializableDjangoJSONDict:
        """Return data to be passed to the rendered JavaScript view.

        This provides a custom label for the view and some extra view-specific
        state.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            A dictionary of options to pass to the view instance.
        """
        label = self.action.get_label(context=context)

        return {
            'label': f'~~{label}~~',
            'specialKey': [123, 456],
        }


class TestActionsRegistry(ActionsRegistry):
    """Empty actions registry for testing purposes.

    Version Added:
        8.0
    """

    def get_defaults(self) -> Iterator[BaseAction]:
        """Return an empty set of defaults.

        Yields:
            reviewboard.actions.base.BaseAction:
            Each action (but none, really).
        """
        yield from []


class TestActionAttachmentPointsRegistry(ActionAttachmentPointsRegistry):
    """Empty actions attachment points registry for testing purposes.

    Version Added:
        8.0
    """

    def get_defaults(self) -> Iterator[ActionAttachmentPoint]:
        """Return an empty set of defaults.

        Yields:
            reviewboard.actions.base.ActionAttachmentPoint:
            Each attachment point (but none, really).
        """
        yield from []



"""A hook for adding new Review UIs."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

from reviewboard.reviews.ui.base import register_ui, unregister_ui


class ReviewUIHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """This hook allows integration of Extension-defined Review UIs.

    This accepts a list of Review UIs specified by the Extension and
    registers them when the hook is created. Likewise, it unregisters
    the same list of Review UIs when the Extension is disabled.
    """

    def initialize(self, review_uis):
        """Initialize the hook.

        This will register the list of review UIs for use in reviewing
        file attachments.

        Args:
            review_uis (list of type):
                The list of review UI classes to register. Each must be a
                subclass of
                :py:class:`~reviewboard.reviews.ui.base.ReviewUI`.

        Raises:
            TypeError:
                The provided review UI class is not of a compatible type.
        """
        self.review_uis = review_uis

        for review_ui in self.review_uis:
            register_ui(review_ui)

    def shutdown(self):
        """Shut down the hook.

        This will unregister the list of review UIs.
        """
        for review_ui in self.review_uis:
            unregister_ui(review_ui)



"""Model administration for SCMTools and repositories."""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from django.contrib import admin
from django.db.models.signals import pre_delete
from django.dispatch import receiver
from django.http import HttpResponse, HttpResponseNotFound
from django.shortcuts import get_object_or_404, render
from django.urls import include, path
from django.utils.html import format_html, mark_safe
from django.utils.translation import gettext, gettext_lazy as _

from reviewboard.accounts.admin import fix_review_counts
from reviewboard.admin import ModelAdmin, admin_site
from reviewboard.admin.server import get_server_url
from reviewboard.hostingsvcs.errors import MissingHostingServiceError
from reviewboard.scmtools.forms import RepositoryForm
from reviewboard.scmtools.models import Repository, Tool

if TYPE_CHECKING:
    from django.utils.safestring import SafeString

    from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
    from reviewboard.scmtools.core import SCMTool


class RepositoryAdmin(ModelAdmin):
    list_display = (
        'name',
        'path',
        '_repository_type',
        '_visible',
        'inline_actions',
    )
    list_select_related = ('hosting_account',)
    search_fields = ('name', 'path', 'mirror_path', 'tool__name')
    raw_id_fields = ('local_site',)
    ordering = ('name',)
    fieldsets = (
        (_('General Information'), {
            'fields': ('name', 'visible',),
            'classes': ('wide',),
        }),
        (RepositoryForm.REPOSITORY_HOSTING_FIELDSET, {
            'fields': (
                'hosting_type',
                'hosting_account',
                'force_authorize',
            ),
            'classes': ('wide',),
        }),
        (RepositoryForm.REPOSITORY_INFO_FIELDSET, {
            'fields': (
                'tool',
                'repository_plan',
            ),
            'classes': ('wide',),
        }),
        (RepositoryForm.SSH_KEY_FIELDSET, {
            'fields': (
                'associate_ssh_key',
            ),
            'classes': ('wide',),
        }),
        (RepositoryForm.BUG_TRACKER_FIELDSET, {
            'fields': (
                'bug_tracker_use_hosting',
                'bug_tracker_type',
                'bug_tracker_hosting_url',
                'bug_tracker_plan',
                'bug_tracker_hosting_account_username',
                'bug_tracker',
            ),
            'classes': ('wide',),
        }),
        (_('Access Control'), {
            'fields': ('public', 'users', 'review_groups'),
            'classes': ('wide',),
        }),
        (_('Advanced Settings'), {
            'fields': ('encoding',),
            'classes': ('wide', 'collapse'),
        }),
        (_('Internal State'), {
            'description': _('<p>This is advanced state that should not be '
                             'modified unless something is wrong.</p>'),
            'fields': ('local_site', 'hooks_uuid', 'extra_data'),
            'classes': ['collapse'],
        }),
    )
    form = RepositoryForm

    fieldset_template_name = 'admin/scmtools/repository/_fieldset.html'

    @admin.display(description=_('Type / Account'))
    def _repository_type(
        self,
        repository: Repository,
    ) -> str:
        result: str
        scmtool: Optional[SCMTool]
        missing_support: bool = False

        try:
            scmtool = repository.get_scmtool()
        except Exception:
            scmtool = None

        if scmtool:
            assert scmtool.name
            scmtool_name = scmtool.name
        else:
            missing_support = True
            scmtool_name = repository.scmtool_id

        hosting_account = repository.hosting_account

        if hosting_account:
            service: BaseHostingService | None

            try:
                service = hosting_account.service
            except MissingHostingServiceError:
                service = None

            if service:
                service_name = hosting_account.service.name
            else:
                service_name = hosting_account.service_name
                missing_support = True

            result = '%s (%s@%s)' % (scmtool_name, hosting_account.username,
                                     service_name)
        else:
            result = scmtool_name

        if missing_support:
            result = gettext('%s (missing support)') % result

        return result

    @admin.display(description='')
    def inline_actions(
        self,
        repository: Repository,
    ) -> SafeString:
        """Return HTML containing actions to show for each repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to return actions for.

        Returns:
            unicode:
            The resulting HTML.
        """
        s = ['<div class="rb-c-admin-change-list__item-actions">']

        def _build_item(
            url: str,
            css_class: str,
            name: str,
        ) -> SafeString:
            css_class = \
                f'rb-c-admin-change-list__item-action ink-c-button {css_class}'

            return format_html('<a class="{0}" href="{1}">{2}</a>',
                               css_class, url, name)

        if repository.hosting_account:
            service: BaseHostingService | None

            try:
                service = repository.hosting_account.service
            except Exception:
                service = None

            if service and service.has_repository_hook_instructions:
                s.append(_build_item(
                    url='%s/hooks-setup/' % repository.pk,
                    css_class='action-hooks-setup',
                    name=_('Hooks')))

        s.append(_build_item(
            url='%s/rbtools-setup/' % repository.pk,
            css_class='action-rbtools-setup',
            name=_('RBTools Setup')))

        s.append('</div>')

        return mark_safe(''.join(s))

    @admin.display(boolean=True,
                   description=_('Show'))
    def _visible(self, repository):
        return repository.visible

    def get_urls(self):
        return [
            path('<int:repository_id>/', include([
                path('hooks-setup/',
                     self.admin_site.admin_view(self.hooks_setup)),
                path('rbtools-setup/',
                     self.admin_site.admin_view(self.rbtools_setup)),
            ])),
        ] + super(RepositoryAdmin, self).get_urls()

    def hooks_setup(self, request, repository_id):
        repository = get_object_or_404(Repository, pk=repository_id)

        if repository.hosting_account:
            service = repository.hosting_account.service

            if service and service.has_repository_hook_instructions:
                return HttpResponse(service.get_repository_hook_instructions(
                    request, repository))

        return HttpResponseNotFound()

    def rbtools_setup(self, request, repository_id):
        repository = get_object_or_404(Repository, pk=repository_id)

        return render(
            request=request,
            template_name='admin/scmtools/repository/rbtools_setup.html',
            context={
                'repository': repository,
                'reviewboard_url': get_server_url(
                    local_site=repository.local_site),
            })


@receiver(pre_delete, sender=Repository,
          dispatch_uid='repository_delete_reset_review_counts')
def repository_delete_reset_review_counts(sender, instance, using, **kwargs):
    """Reset review counts in the dashboard when deleting repository objects.

    There doesn't seem to be a good way to get notified on cascaded delete
    operations, which means that when deleting a repository, there's no
    good way to update the review counts that are shown to users. This
    method clears them out entirely to be regenerated. Deleting
    repositories should be a very rare occurrence, so it's not too
    upsetting to do this.
    """
    fix_review_counts()


class ToolAdmin(ModelAdmin):
    list_display = ('__str__', 'class_name')


admin_site.register(Repository, RepositoryAdmin)
admin_site.register(Tool, ToolAdmin)



"""A hook for determining if a review request is approved."""

from __future__ import annotations

from typing import TYPE_CHECKING

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

if TYPE_CHECKING:
    from reviewboard.reviews.models import ReviewRequest


class ReviewRequestApprovalHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for determining if a review request is approved.

    Extensions can use this to hook into the process for determining
    review request approval, which may impact any scripts integrating
    with Review Board to, for example, allow committing to a repository.

    .. seealso::

        * :ref:`ReviewRequestApprovalHook Developer Guide
          <review-request-approval-hook>`
    """

    def is_approved(
        self,
        review_request: ReviewRequest,
        prev_approved: bool,
        prev_failure: str | None,
    ) -> bool | tuple[bool, str | None]:
        """Determine if the review request is approved.

        This function is provided with the review request and the previously
        calculated approved state (either from a prior hook, or from the
        base state of ``ship_it_count > 0 and issue_open_count == 0``).

        If approved, this should return True. If unapproved, it should
        return a tuple with False and a string briefly explaining why it's
        not approved. This may be displayed to the user.

        It generally should also take the previous approved state into
        consideration in this choice (such as returning False if the previous
        state is False). This is, however, fully up to the hook.

        The approval decision may be overridden by any following hooks.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being checked for approval.

            prev_approved (bool):
                The previously-calculated approval result, either from another
                hook or by Review Board.

            prev_failure (str):
                The previously-calculated approval failure message, either
                from another hook or by Review Board.

        Returns:
            bool or tuple:
            Either a boolean indicating approval (re-using ``prev_failure``,
            if not approved), or a tuple in the form of
            ``(approved, failure_message)``.
        """
        raise NotImplementedError



import logging

from django.db.models import Q
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_response_errors,
                                       webapi_request_fields)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_FORM_DATA,
                                   NOT_LOGGED_IN, PERMISSION_DENIED)
from djblets.webapi.fields import FileFieldType, StringFieldType

from reviewboard.certs.errors import CertificateVerificationError
from reviewboard.diffviewer.errors import (DiffParserError,
                                           DiffTooBigError,
                                           EmptyDiffError)
from reviewboard.diffviewer.models import DiffSet
from reviewboard.scmtools.models import Repository
from reviewboard.scmtools.errors import (FileNotFoundError,
                                         SCMError)
from reviewboard.scmtools.git import ShortSHA1Error
from reviewboard.webapi.decorators import (webapi_check_login_required,
                                           webapi_check_local_site)
from reviewboard.webapi.errors import (DIFF_EMPTY,
                                       DIFF_PARSE_ERROR,
                                       DIFF_TOO_BIG,
                                       INVALID_REPOSITORY,
                                       REPO_FILE_NOT_FOUND,
                                       UNVERIFIED_HOST_CERT)
from reviewboard.webapi.resources.diff import DiffResource


logger = logging.getLogger(__name__)


class ValidateDiffResource(DiffResource):
    """Verifies whether a diff file will work.

    This allows clients to validate whether a diff file (with optional parent
    diff) can be parsed and displayed, without actually creating a review
    request first.
    """
    added_in = '2.0'

    singleton = True
    name = 'diff_validation'
    uri_name = 'diffs'
    uri_object_key = None

    allowed_methods = ('GET', 'POST',)

    item_child_resources = []
    list_child_resources = []

    @webapi_check_local_site
    @webapi_check_login_required
    def get(self, request, *args, **kwargs):
        """Returns links for using this resource."""
        return 200, {
            'links': self.get_links(request=request, *args, **kwargs),
        }

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(
        DIFF_EMPTY,
        DIFF_PARSE_ERROR,
        DIFF_TOO_BIG,
        DOES_NOT_EXIST,
        INVALID_FORM_DATA,
        INVALID_REPOSITORY,
        NOT_LOGGED_IN,
        PERMISSION_DENIED,
        REPO_FILE_NOT_FOUND,
        UNVERIFIED_HOST_CERT,
    )
    @webapi_request_fields(
        required={
            'repository': {
                'type': StringFieldType,
                'description': 'The path or ID of the repository.',
            },
            'path': {
                'type': FileFieldType,
                'description': 'The main diff file.',
            },
        },
        optional={
            'basedir': {
                'type': StringFieldType,
                'description': 'The base directory that will prepended to '
                               'all paths in the diff. This is needed for '
                               'some types of repositories. The directory '
                               'must be between the root of the repository '
                               'and the top directory referenced in the '
                               'diff paths.',
            },
            'parent_diff_path': {
                'type': FileFieldType,
                'description': 'The optional parent diff to upload.',
            },
            'base_commit_id': {
                'type': StringFieldType,
                'description': 'The ID/revision this change is built upon. '
                               'If using a parent diff, then this is the base '
                               'for that diff. This may not be provided for '
                               'all diffs or repository types, depending on '
                               'how the diff was uploaded.',
            },
        }
    )
    def create(self, request, repository, basedir=None, local_site_name=None,
               base_commit_id=None, *args, **kwargs):
        """Validate a diff.

        This API has a similar signature to the ReviewRequest resource POST
        API, but instead of actually creating a review request, will return
        either OK or an error, depending on whether the included diff file
        parsed correctly.
        """
        local_site = self._get_local_site(local_site_name)

        path = request.FILES.get('path')
        parent_diff_path = request.FILES.get('parent_diff_path')

        try:
            q = Q(pk=int(repository))
        except ValueError:
            q = (Q(path=repository) |
                 Q(mirror_path=repository) |
                 Q(name=repository))

        repositories = (
            Repository.objects
            .accessible(request.user, local_site=local_site)
            .filter(q)
        )
        repository_count = repositories.count()

        if repository_count == 0:
            return INVALID_REPOSITORY, {
                'repository': repository,
            }
        elif repository_count > 1:
            msg = ('Too many repositories matched "%s". '
                   'Try specifying the repository by name instead.'
                   % repository)

            return INVALID_REPOSITORY.with_message(msg), {
                'repository': repository,
            }

        repository = repositories.first()

        if not repository.diffs_use_absolute_paths and basedir is None:
            return INVALID_FORM_DATA, {
                'fields': {
                    'basedir': ['Given repository requires a base directory'],
                },
            }

        if basedir is None:
            # If we get here, the repository uses absolute paths. Deeper down
            # (where we don't necessarily know about the details of the
            # repository), we do an os.path.join() with the basedir value,
            # which will choke if it's None.
            basedir = ''

        try:
            DiffSet.objects.create_from_upload(
                repository=repository,
                diff_file=path,
                parent_diff_file=parent_diff_path,
                basedir=basedir,
                request=request,
                base_commit_id=base_commit_id,
                validate_only=True)
        except FileNotFoundError as e:
            return REPO_FILE_NOT_FOUND, {
                'file': e.path,
                'revision': str(e.revision),
            }
        except EmptyDiffError:
            return DIFF_EMPTY
        except DiffTooBigError as e:
            return DIFF_TOO_BIG, {
                'reason': str(e),
                'max_size': e.max_diff_size,
            }
        except DiffParserError as e:
            return DIFF_PARSE_ERROR, {
                'reason': str(e),
                'linenum': e.linenum,
            }
        except ShortSHA1Error as e:
            return REPO_FILE_NOT_FOUND, {
                'reason': str(e),
                'file': e.path,
                'revision': str(e.revision),
            }
        except CertificateVerificationError as e:
            return UNVERIFIED_HOST_CERT.with_message(str(e)), {
                'certificate': e.certificate,
            }
        except SCMError as e:
            return DIFF_PARSE_ERROR.with_message(str(e))
        except Exception as e:
            logger.exception('Unexpected error when validating diff.')

            return DIFF_PARSE_ERROR.with_message(
                'Unexpected error when validating the diff: %s' % e)

        return 200, {}

    def _build_named_url(self, name):
        """Builds a Django URL name from the provided name."""
        return 'validate-diffs-resource'


validate_diff_resource = ValidateDiffResource()



"""Hosting service for Review Board Gateway."""

from __future__ import annotations

import hashlib
import hmac
import json
import logging
from collections import defaultdict
from typing import TYPE_CHECKING
from urllib.error import HTTPError
from urllib.parse import quote

from django import forms
from django.http import HttpResponse, HttpResponseBadRequest
from django.template.loader import render_to_string
from django.urls import path
from django.utils.translation import gettext_lazy as _, gettext
from housekeeping import deprecate_non_keyword_only_args

from reviewboard.admin.server import build_server_url, get_server_url
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base.client import HostingServiceClient
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceAPIError,
                                            HostingServiceError)
from reviewboard.hostingsvcs.hook_utils import (close_all_review_requests,
                                                get_repository_for_hook,
                                                get_review_request_id)
from reviewboard.scmtools.core import Branch, Commit, UNKNOWN
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import (FileNotFoundError,
                                         RepositoryNotFoundError)
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


def hook_close_submitted(
    request: HttpRequest,
    local_site_name: (str | None) = None,
    repository_id: (int | None) = None,
    hosting_service_id: (str | None) = None,
) -> HttpResponse:
    """Close review requests as submitted after a push.

    Args:
        request (django.http.HttpRequest):
            The request from the RB Gateway webhook.

        local_site_name (str, optional):
            The local site name, if available.

        repository_id (int, optional):
            The ID of the repository, if available.

        hosting_service_id (str, optional):
            The ID of the hosting service.

    Returns:
        django.http.HttpResponse;
        A response for the request.
    """
    hook_event = request.META.get('HTTP_X_RBG_EVENT')

    if hook_event == 'ping':
        return HttpResponse()
    elif hook_event != 'push':
        return HttpResponseBadRequest(
            'Only "ping" and "push" events are supported.')

    assert repository_id is not None
    assert hosting_service_id is not None

    repository = get_repository_for_hook(
        repository_id=repository_id,
        hosting_service_id=hosting_service_id,
        local_site_name=local_site_name)

    sig = request.META.get('HTTP_X_RBG_SIGNATURE', '')
    m = hmac.new(repository.get_or_create_hooks_uuid().encode('utf-8'),
                 request.body,
                 hashlib.sha1)

    if not hmac.compare_digest(m.hexdigest(), sig):
        return HttpResponseBadRequest('Bad signature.')

    try:
        payload = json.loads(request.body.decode('utf-8'))
    except ValueError as e:
        logger.error('The payload is not in JSON format: %s', e,
                     extra={'request': request})
        return HttpResponseBadRequest('Invalid payload format.')

    if 'commits' not in payload:
        return HttpResponseBadRequest('Invalid payload; expected "commits".')

    server_url = get_server_url(request=request)
    review_request_ids_to_commits = defaultdict(list)

    for commit in payload['commits']:
        commit_id = commit.get('id')
        commit_message = commit.get('message')
        review_request_id = get_review_request_id(
            commit_message=commit_message,
            server_url=server_url,
            commit_id=commit_id,
            repository=repository)

        targets = commit['target']

        if 'tags' in targets and targets['tags']:
            target = targets['tags'][0]
        elif 'bookmarks' in targets and targets['bookmarks']:
            target = targets['bookmarks'][0]
        elif 'branch' in targets:
            target = targets['branch']
        else:
            target = ''

        if target:
            target_str = '%s (%s)' % (target, commit_id[:7])
        else:
            target_str = commit_id[:7]

        review_request_ids_to_commits[review_request_id].append(target_str)

    if review_request_ids_to_commits:
        close_all_review_requests(
            review_request_id_to_commits=review_request_ids_to_commits,
            local_site_name=local_site_name,
            repository=repository,
            hosting_service_id=hosting_service_id)

    return HttpResponse()


class ReviewBoardGatewayForm(BaseHostingServiceRepositoryForm):
    """Hosting service form for Review Board Gateway.

    Provide an additional field on top of the base hosting service form to
    allow specification of the repository name.
    """

    rbgateway_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=128,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the name '
                    'specified in the configuration file for rb-gateway.'))


class ReviewBoardGatewayClient(HostingServiceClient):
    """Client interface to the RB Gateway API."""

    def __init__(self, *args, **kwargs):
        """Initialize the client.

        Args:
            *args (tuple):
                Positional arguments for the parent class.

            **kwargs (dict):
                Keyword arguments for the parent class.
        """
        super(ReviewBoardGatewayClient, self).__init__(*args, **kwargs)

        self.api_url = self.hosting_service.account.hosting_url

    def api_authenticate(self, username, password):
        """Authenticate against the RB Gateway server.

        This will attempt to authenticate with the given credentials. If
        successful, information on the session, including an API token for
        further API requests, will be returned.

        Args:
            username (unicode):
                The username to authenticate with.

            password (unicode):
                The password to authenticate with.

        Returns:
            dict:
            The new session information.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file contents. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        try:
            response = self.http_post('%s/session' % self.api_url,
                                      username=username,
                                      password=password)

            return response.json
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                raise HostingServiceAPIError(
                    gettext('A Review Board Gateway server was not found at '
                            'the provided URL. Make sure you are providing '
                            'the root of the server, and not a path '
                            'within it.'))

            raise

    def api_get_repository(self, repo_name):
        """Return information on a repository.

        Args:
            repo_name (unicode):
                The name of the repository.

        Returns:
            dict:
            The repository information.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file contents. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = '%s/path' % self._get_repos_api_url(repo_name)

        return self.http_get(url).json

    def api_get_file_contents(self, repo_name, revision, base_commit_id, path):
        """Return a file from a repository.

        This will perform an API request to fetch the contents of a file.

        Args:
            repo_name (unicode):
                The name of the repository registered on RB Gateway.

            path (unicode):
                The file path.

            revision (unicode):
                The revision of the file to retrieve.

            base_commit_id (unicode, optional):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file contents. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.FileNotFoundError:
                The file was not found.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = self._get_file_api_url(repo_name=repo_name,
                                     revision=revision,
                                     base_commit_id=base_commit_id,
                                     path=path)

        try:
            return self.http_get(url).data
        except HostingServiceError as e:
            if e.http_code == 404:
                raise FileNotFoundError(path, revision)

            raise

    def api_get_file_exists(self, repo_name, revision, base_commit_id,
                            path):
        """Return whether a file exists in a repository.

        This will perform an API request to fetch information on the file,
        using that to determine if the file exists.

        Args:
            repo_name (unicode):
                The name of the repository registered on RB Gateway.

            path (unicode):
                The file path.

            revision (unicode):
                The revision of the file.

            base_commit_id (unicode, optional):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

        Returns:
            bool:
            ``True`` if the file exists. ``False`` if it does not.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the file information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = self._get_file_api_url(repo_name=repo_name,
                                     revision=revision,
                                     base_commit_id=base_commit_id,
                                     path=path)

        try:
            self.http_head(url)
            return True
        except HostingServiceError as e:
            if e.http_code == 404:
                return False

            raise

    def api_get_branches(self, repo_name):
        """Return a list of branches for a repository.

        Args:
            repo_name (unicode):
                The name of the repository.

        Returns:
            list of dict:
            The list of branches in the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.
        """
        url = self._get_branches_api_url(repo_name)

        return self.http_get(url).json

    def api_get_commits(self, repo_name, branch_name, start=None):
        """Return a list of commits for a repository.

        Args:
            repo_name (unicode):
                The name of the repository.

            branch_name (unicode):
                The name of the branch to list commits on.

            start (unicode, optional):
                The optional starting commit ID for the list. This is used
                for pagination purposes.

        Returns:
            list of dict:
            The list of commits in the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = self._get_commits_api_url(repo_name, branch_name=branch_name)

        if start is not None:
            url = '%s?start=%s' % (url, start)

        return self.http_get(url).json

    def api_get_commit(self, repo_name, commit_id):
        """Return a commit at a given revision/commit ID.

        Args:
            repo_name (unicode):
                The name of the repository.

            revision (unicode):
                The revision/ID of the commit to fetch.

        Returns:
            dict:
            Information on the commit from the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised. See :py:meth:`process_http_error`.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        url = self._get_commits_api_url(repo_name, commit_id)

        return self.http_get(url).json

    def get_http_credentials(self, account, username=None, password=None,
                             **kwargs):
        """Return credentials used to authenticate with RB Gateway.

        If a username and password is provided, this will authenticate using
        HTTP Basic Auth. This is needed for initially linking an account or
        updating credentials.

        If instead the account data contains a private token from a previous
        authentication request, it will be provided to RB Gateway through the
        :mailheader:`PRIVATE-TOKEN` header.

        Args:
            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The stored authentication data for the account.

            username (unicode, optional):
                An explicit username passed by the caller. This will override
                the token stored in the account, if both a username and
                password are provided.

            password (unicode, optional):
                An explicit password passed by the caller. This will override
                the token stored in the account, if both a username and
                password are provided.

            **kwargs (dict, unused):
                Additional keyword arguments passed in when making the HTTP
                request.

        Returns:
            dict:
            A dictionary of credentials for the request.
        """
        if username is None and password is None:
            private_token = self.hosting_service.get_private_token()

            if private_token is not None:
                return {
                    'headers': {
                        'PRIVATE-TOKEN': private_token,
                    },
                }

        return super(ReviewBoardGatewayClient, self).get_http_credentials(
            account=account,
            username=username,
            password=password,
            **kwargs)

    def process_http_error(self, request, e):
        """Process an HTTP error, raising a result.

        This will look at the error, raising a more suitable exception
        in its place.

        Args:
            request (reviewboard.hostingsvcs.base.http.
                     HostingServiceHTTPRequest, unused):
                The request that resulted in an error.

            e (urllib2.URLError):
                The error to check.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.

            reviewboard.hostingsvcs.errors.HostingServiceAPIError:
                An error occurred communicating with the API. An unparsed
                payload is available.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an unexpected error performing the request.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        # Perform any default checks.
        super(ReviewBoardGatewayClient, self).process_http_error(request, e)

        if isinstance(e, HTTPError):
            code = e.getcode()

            if e.code == 401:
                raise AuthorizationError(
                    gettext('The username or password is incorrect.'))
            elif e.code == 404:
                raise HostingServiceAPIError(
                    gettext('The API endpoint was not found.'),
                    http_code=code)
            else:
                msg = e.read()

                raise HostingServiceAPIError(msg,
                                             http_code=code,
                                             rsp=msg)
        else:
            raise HostingServiceError(e.reason)

    def _get_repos_api_url(self, repo_name=None):
        """Return the API URL for working with repositories.

        Args:
            repo_name (unicode, optional):
                The optional name of a repository to include in the URL.

        Returns:
            unicode:
            The URL for working with a list of repositories or the specified
            repository.
        """
        url = '%s/repos' % self.api_url

        if repo_name is not None:
            url = '%s/%s' % (url, quote(repo_name))

        return url

    def _get_branches_api_url(self, repo_name, branch_name=None):
        """Return the API URL for branches on a repository.

        Args:
            repo_name (unicode):
                The name of the repository.

            branch_name (unicode, optional):
                The optional name of a branch to include in the URL.

        Returns:
            unicode:
            The URL for working with a list of branches or a specific branch.
        """
        url = '%s/branches' % self._get_repos_api_url(repo_name)

        if branch_name is not None:
            url = '%s/%s' % (url, quote(branch_name))

        return url

    def _get_commits_api_url(self, repo_name, commit_id=None,
                             branch_name=None):
        """Return the API URL for commits on a repository.

        Args:
            repo_name (unicode):
                The name of the repository.

            commit_id (unicode, optional):
                The optional commit ID to include in the URL.

            branch_name (unicode, optional):
                The optional branch to fetch commits from.

        Returns:
            unicode:
            The URL for working with a list of commits or a specific commit.
        """
        if branch_name is None or commit_id is not None:
            url = '%s/commits' % self._get_repos_api_url(repo_name)
        else:
            url = '%s/commits' % self._get_branches_api_url(repo_name,
                                                            branch_name)

        if commit_id is not None:
            url = '%s/%s' % (url, quote(commit_id))

        return url

    def _get_file_api_url(self, repo_name, revision, base_commit_id=None,
                          path=None):
        """Return the URL for accessing information about a file.

        A revision or a (base commit ID, path) pair is expected to be provided.
        By default, this will return the URL based on the revision, if both
        are provided.

        Args:
            repo_name (unicode):
                The name of the repository registered on RB Gateway.

            revision (unicode):
                The revision of the file to retrieve.

            base_commit_id (unicode, optional):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

            path (unicode, optional):
                The file path.

        Returns:
            unicode:
            The URL for fetching file information.
        """
        if revision and revision != UNKNOWN:
            return ('%s/file/%s'
                    % (self._get_repos_api_url(repo_name), quote(revision)))
        else:
            assert path is not None
            return ('%s/path/%s'
                    % (self._get_commits_api_url(repo_name, base_commit_id),
                       quote(path)))


class ReviewBoardGateway(BaseHostingService):
    """Hosting service support for Review Board Gateway.

    Review Board Gateway is a lightweight self-installed source hosting service
    that provides an API around self-hosted source code repositories.

    More information can be found at
    https://www.reviewboard.org/downloads/rbgateway/
    """

    name = 'Review Board Gateway'
    hosting_service_id = 'rbgateway'

    client_class = ReviewBoardGatewayClient
    form = ReviewBoardGatewayForm

    self_hosted = True
    needs_authorization = True
    supports_repositories = True
    supports_post_commit = True
    supported_scmtools = ['Git', 'Mercurial']

    has_repository_hook_instructions = True

    repository_fields = {
        'Git': {
            'path': '%(hosting_url)s/repos/%(rbgateway_repo_name)s/path',
        },
        'Mercurial': {
            'path': '%(hosting_url)s/repos/%(rbgateway_repo_name)s/path',
        },
    }

    repository_url_patterns = [
        path('hooks/close-submitted/',
             hook_close_submitted,
             name='rbgateway-hooks-close-submitted'),
    ]

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def check_repository(
        self,
        *,
        rbgateway_repo_name: str,
        **kwargs,
    ) -> None:
        """Checks the validity of a repository configuration.

        Args:
            rbgateway_repo_name (str):
                The name of the repository to check.

            **kwargs (dict, unused):
                Additional keyword arguments provided by the repository form.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                The repository is not valid.

            reviewboard.scmtools.errors.RepositoryNotFoundError:
                The repository was not found.
        """
        try:
            self.client.api_get_repository(rbgateway_repo_name)
        except HostingServiceAPIError as e:
            if e.http_code == 404:
                raise RepositoryNotFoundError()

            raise

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def authorize(
        self,
        *,
        username: str | None,
        password: str | None,
        **kwargs,
    ) -> None:
        """Authorize an account on the RB Gateway service.

        This will perform an authentication request against the API. If
        successful, the generated API token will be stored, encrypted, for
        future requests to the API.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            username (str):
                The username for the account.

            password (str):
                The password for the account.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        auth_data = self.client.api_authenticate(username, password)

        self.account.data['private_token'] = \
            encrypt_password(auth_data['private_token'])
        self.account.save()

    def is_authorized(self):
        """Return if the account has a stored auth token.

        This will check if we have previously stored a private token for the
        account. It does not validate that the token still works.
        """
        return 'private_token' in self.account.data

    def get_private_token(self):
        """Return the private token used for authentication.

        Returns:
            unicode:
            The private token, or ``None`` if one wasn't set.
        """
        private_token = self.account.data.get('private_token')

        if not private_token:
            return None

        return decrypt_password(private_token)

    def get_file(self, repository, path, revision, base_commit_id, *args,
                 **kwargs):
        """Return a file from a repository.

        This will perform an API request to fetch the contents of a file.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve the file from.

            path (unicode):
                The file path.

            revision (unicode):
                The revision the file should be retrieved from.

            base_commit_id (unicode):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.

            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found.
        """
        return self.client.api_get_file_contents(
            repo_name=self._get_repo_name(repository),
            revision=revision,
            base_commit_id=base_commit_id,
            path=path)

    def get_file_exists(self, repository, path, revision, base_commit_id,
                        *args, **kwargs):
        """Return whether a file exists in a repository.

        This will perform an API request to fetch information on the file,
        using that to determine if the file exists.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve the file from.

            path (unicode):
                The file path.

            revision (unicode):
                The revision the file should be retrieved from.

            base_commit_id (unicode):
                The ID of the commit that the file was changed in. This may
                not be provided, and is dependent on the type of repository.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            bool:
            ``True`` if the file exists. ``False`` if it does not.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        return self.client.api_get_file_exists(
            repo_name=self._get_repo_name(repository),
            revision=revision,
            base_commit_id=base_commit_id,
            path=path)

    def get_branches(self, repository):
        """Return the branches for the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch branches for.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The branches returned for the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        repo_name = self._get_repo_name(repository)
        tool_name = repository.scmtool_class.name

        if tool_name == 'Git':
            default_branch = 'master'
        elif tool_name == 'Mercurial':
            default_branch = 'default'
        else:
            default_branch = None

        return [
            Branch(id=branch_info['name'],
                   commit=branch_info['id'],
                   default=(branch_info['name'] == default_branch))
            for branch_info in self.client.api_get_branches(repo_name)
        ]

    def get_commits(self, repository, branch, start=None):
        """Return a list of commits for a repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch commits on.

            branch (unicode):
                The name of the branch to list commits on.

            start (unicode, optional):
                The optional starting commit ID for the list. This is used
                for pagination purposes.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The list of commits in the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        repo_name = self._get_repo_name(repository)

        return [
            Commit(author_name=commit_info['author'],
                   id=commit_info['id'],
                   date=commit_info['date'],
                   message=commit_info['message'],
                   parent=commit_info['parent_id'])
            for commit_info in self.client.api_get_commits(repo_name, branch)
        ]

    def get_change(self, repository, revision):
        """Return a commit at a given revision/commit ID.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch the commit from.

            revision (unicode):
                The revision/ID of the commit to fetch.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit from the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                Error retrieving the branch information. There may be a more
                specific subclass raised.
        """
        repo_name = self._get_repo_name(repository)
        commit_info = self.client.api_get_commit(repo_name, revision)

        return Commit(author_name=commit_info['author'],
                      id=commit_info['id'],
                      date=commit_info['date'],
                      message=commit_info['message'],
                      parent=commit_info['parent_id'],
                      diff=commit_info['diff'].encode('utf-8'))

    def get_repository_hook_instructions(self, request, repository):
        """Returns instructions for setting up incoming webhooks.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            repository (reviewboard.scmtools.models.Repository):
                The repository to show instructions for.

        Returns:
            django.http.HttpResponse:
            The hook installation instructions rendered to the contents of an
            HTTP response.
        """
        example_id = 123
        example_url = build_server_url(local_site_reverse(
            'review-request-detail',
            local_site=repository.local_site,
            kwargs={
                'review_request_id': example_id,
            }))

        hook_uuid = repository.get_or_create_hooks_uuid()
        close_url = build_server_url(local_site_reverse(
            'rbgateway-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            },
            local_site=repository.local_site))

        return render_to_string(
            template_name='hostingsvcs/rb-gateway/repo_hook_instructions.html',
            request=request,
            context={
                'example_id': example_id,
                'example_url': example_url,
                'hook_uuid': hook_uuid,
                'close_url': close_url,
                'repo_name': repository.extra_data['rbgateway_repo_name'],
            })

    def _get_repo_name(self, repository):
        """Return the stored API name for the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to return the API name for.

        Returns:
            unicode:
            The API name for the repository.
        """
        return repository.extra_data['rbgateway_repo_name']



"""Administration UI support for integration configurations."""

from reviewboard.admin import ModelAdmin, admin_site
from reviewboard.integrations.models import IntegrationConfig


class IntegrationConfigAdmin(ModelAdmin):
    list_display = ('integration_id', 'name', 'enabled', 'last_updated')
    raw_id_fields = ('local_site',)


admin_site.register(IntegrationConfig, IntegrationConfigAdmin)



"""Testing utilities for diff parsers.

Version Added:
    4.0.6
"""


class DiffParserTestingMixin(object):
    """Testing utilities for diff parsers.

    This must be mixed into a class using
    :py:class:`djblets.testing.testcases.TestCase`.

    Version Added:
        4.0.6
    """

    def assert_parsed_diff(self, parsed_diff, num_changes=0, **attrs):
        """Assert that a ParsedDiff has the expected attributes.

        Args:
            parsed_diff (reviewboard.diffviewer.parser.ParsedDiff):
                The parsed diff to check.

            num_changed (int, optional):
                The expected number of changes in the parsed diff.

            **attrs (dict, optional):
                A dictionary of explicit non-default attribute values to check.

        Raises:
            AssertionError:
                One or more attributes do not match.
        """
        self._assert_parsed_diff_obj(
            parsed_diff,
            attrs=attrs,
            default_attrs={
                'uses_commit_ids_as_revisions': False,
            })
        self.assertEqual(len(parsed_diff.changes), num_changes)

    def assert_parsed_diff_change(self, parsed_diff_change, num_files=0,
                                  **attrs):
        """Assert that a ParsedDiffChange has the expected attributes.

        Args:
            parsed_diff (reviewboard.diffviewer.parser.ParsedDiffChange):
                The parsed diff change to check.

            num_files (int, optional):
                The expected number of files in the parsed change.

            **attrs (dict, optional):
                A dictionary of explicit non-default attribute values to check.

        Raises:
            AssertionError:
                One or more attributes do not match.
        """
        self._assert_parsed_diff_obj(
            parsed_diff_change,
            attrs=attrs,
            default_attrs={
                'commit_id': None,
                'parent_commit_id': None,
            })
        self.assertEqual(len(parsed_diff_change.files), num_files)

    def assert_parsed_diff_file(self, parsed_diff_file, **attrs):
        """Assert that a ParsedDiffFile has the expected attributes.

        Args:
            parsed_diff_file (reviewboard.diffviewer.parser.ParsedDiffFile):
                The parsed diff file to check.

            **attrs (dict, optional):
                A dictionary of explicit non-default attribute values to check.

        Raises:
            AssertionError:
                One or more attributes do not match.
        """
        data = attrs.pop('data', b'')

        self._assert_parsed_diff_obj(
            parsed_diff_file,
            attrs=attrs,
            default_attrs={
                'binary': False,
                'copied': False,
                'delete_count': 0,
                'deleted': False,
                'index_header_value': None,
                'insert_count': 0,
                'is_symlink': False,
                'moved': False,
                'new_symlink_target': None,
                'new_unix_mode': None,
                'old_symlink_target': None,
                'old_unix_mode': None,
            })

        self.assertEqual(parsed_diff_file.data.splitlines(),
                         data.splitlines())

    def _assert_parsed_diff_obj(self, parsed_diff_obj, attrs, default_attrs):
        """Utility to assert that a parsed object has the expected attributes.

        Args:
            parsed_diff_obj (object):
                The parsed diff object to check.

            attrs (dict):
                A dictionary of explicit attribute values to check.

            default_attrs (dict):
                Default attribute values to check, if not explicitly provided.

        Raises:
            AssertionError:
                One or more attributes do not match.
        """
        # We'll test this separately in order to get a better error when
        # assertions fail on this.
        extra_data = attrs.pop('extra_data', {})

        self.assertAttrsEqual(parsed_diff_obj, dict(default_attrs, **attrs))
        self.assertEqual(parsed_diff_obj.extra_data, extra_data)



"""Base HTTP support for hosting services.

Version Added:
    6.0:
    This replaces the HTTP code in the old
    :py:mod:`reviewboard.hostingsvcs.service` module.
"""

from __future__ import annotations

import base64
import json
import logging
from collections import OrderedDict
from typing import Any, Dict, TypedDict, TYPE_CHECKING, Union
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse
from urllib.request import (
    Request as BaseURLRequest,
    HTTPBasicAuthHandler,
    HTTPDigestAuthHandler,
    HTTPPasswordMgrWithDefaultRealm,
    build_opener)

from django.utils.encoding import force_str
from djblets.log import log_timed
from djblets.util.decorators import cached_property

from reviewboard.certs.cert import Certificate
from reviewboard.certs.http import CertificateVerificationHTTPSHandler
from reviewboard.certs.manager import cert_manager
from reviewboard.deprecation import RemovedInReviewBoard90Warning

if TYPE_CHECKING:
    from urllib.request import BaseHandler

    from typelets.funcs import KwargsDict
    from typelets.json import JSONValue
    from typing_extensions import Never, TypeAlias

    from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
    from reviewboard.hostingsvcs.models import HostingServiceAccount


logger = logging.getLogger(__name__)


class UploadedFileInfo(TypedDict):
    """Information on an uploaded file.

    Version Added:
        6.0
    """

    #: The content of the file.
    #:
    #: Type:
    #:      bytes
    content: bytes | str

    #: The base name of the file.
    #:
    #: Type:
    #:      bytes
    filename: bytes | str


#: Form field data to set when performing a POST or PUT.
#:
#: Version Added:
#:     6.0
FormFields: TypeAlias = Dict[Union[bytes, str],
                             Union[bytes, str]]


#: Type for a mapping of HTTP headers for a request or response.
#:
#: Version Added:
#:     6.0
HTTPHeaders: TypeAlias = Dict[str, str]


#: Query arguments for a HTTP request.
#:
#: Version Added:
#:     6.0
QueryArgs: TypeAlias = Dict[str, Any]


#: Form field data to set when performing a POST or PUT.
#:
#: Version Added:
#:     6.0
UploadedFiles: TypeAlias = Dict[Union[bytes, str], UploadedFileInfo]


def _build_ssl_kwargs_from_ssl_cert(
    *,
    hosting_account: HostingServiceAccount,
) -> KwargsDict:
    """Return keyword arguments for an HTTPS handler with legacy cert data.

    This will attempt to migrate the stored ``ssl_cert`` data a hosting
    service account to a modern stored certificate that can then be
    properly managed.

    If there are any issues with migration, or if the resulting SSL
    certificate does not match the target hostname, then the stored certificate
    data will be used directly with hostname verification disabled.

    This should never be needed for hosting service accounts configured
    after Review Board 8.0.

    Version Added:
        8.0

    Args:
        hosting_account (reviewboard.hostingsvcs.models.HostingServiceAccount):
            The hosting service account to migrate.

    Returns:
        dict:
        Keyword arguments to pass to :py:class:`reviewboard.certs.http.
        CertificateVerificationHTTPSHandler`.
    """
    cert_data = hosting_account.data['ssl_cert']
    migrated = True

    if hosting_url := hosting_account.hosting_url:
        local_site = hosting_account.local_site

        try:
            parsed = urlparse(hosting_url)

            if hostname := parsed.hostname:
                port = parsed.port or 443

                # Check if there's an existing certificate being managed.
                certificate = cert_manager.get_certificate(
                    hostname=hostname,
                    port=port,
                    local_site=local_site,
                )

                if certificate is None:
                    # There's no existing certificate. Generate one from
                    # the stored data and check to see if it matches the
                    # hostname. If not, we can't add it, and instead need
                    # to go the legacy `check_hostname=False` route.
                    certificate = Certificate(
                        hostname=hostname,
                        port=port,
                        cert_data=cert_data.encode('ascii'),
                    )

                    if certificate.matches_host(hostname):
                        # This is a direct match, so add this to the cert
                        # manager.
                        cert_manager.add_certificate(
                            certificate=certificate,
                            local_site=local_site,
                        )

                        del hosting_account.data['ssl_cert']
                        hosting_account.save(update_fields=('data',))
                    else:
                        # This is NOT a direct match, but the admin had
                        # previously approved this cert for this server.
                        # We'll have to keep the legacy fallback that
                        # disables hostname checks.
                        logger.warning(
                            'The approved SSL/TLS certificate stored in '
                            'hosting service account ID=%r does not match the '
                            'hostname %r for the server. Falling back to a '
                            'less-secure form of verification. Please ensure '
                            'the server has a valid certificate matching its '
                            'hostname and then upload a new certificate.',
                            hosting_account.pk, hostname,
                        )

                        certificate = None

                if certificate is not None:
                    # A certificate was parsed and valid, so we don't have
                    # to fall back to legacy behavior below.
                    migrated = False
        except Exception:
            # This will be issued every time this cert is used, making it
            # loud and noisy in order to better catch an admin's attention.
            logger.exception(
                'Unexpected error migrating legacy ssl_cert data for hosting '
                'service account %s. Falling back to a legacy insecure '
                'SSL context.',
                hosting_account.pk,
            )

    if migrated:
        # Fall back to using the certificate data directly without hostname
        # validation.
        return {
            'check_hostname': False,
            'extra_cert_data': cert_data,
        }

    # No legacy arguments are required. The certificate was successfully
    # migrated to cert storage.
    return {}


def _log_and_raise(
    value: Never,
    request: HostingServiceHTTPRequest,
    msg: str,
    **fmt_dict,
) -> Never:
    """Log and raise an exception with the given message.

    This is used when validating data going into the request, and is
    intended to help with debugging bad calls to the HTTP code.

    Version Changed:
        8.0:
        Modified to take in the offending value.

    Version Added:
        4.0

    Args:
        value (object):
            The value which was not valid.

        request (HostingServiceHTTPRequest):
            The HTTP request.

        msg (str):
            The error message as a format string.

        **fmt_dict (dict):
            Values for the error message's format string.

    Raises:
        TypeError:
            The exception containing the provided message.
    """
    msg %= dict({
        'method': request.method,
        'service': type(request.hosting_service),
    }, **fmt_dict)

    logger.error(msg)

    raise TypeError(msg)


class HostingServiceHTTPRequest:
    """A request that can use any HTTP method.

    This provides some additional type checking and utilities for working
    with HTTP requests, headers, URL openers, and SSL certification management.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.service` to
          :py:mod:`reviewboard.hostingsvcs.base.http`.

    Version Added:
        4.0
    """

    ######################
    # Instance variables #
    ######################

    #: The request payload body, if any.
    #:
    #: Type:
    #:     bytes
    body: bytes | None

    #: The headers to send in the request.
    #:
    #: Type:
    #:     dict
    headers: HTTPHeaders

    #: The hosting service this request is associated with.
    #:
    #: Type:
    #:     HostingService
    hosting_service: BaseHostingService | None

    #: The HTTP method to perform.
    #:
    #: Type:
    #:     str
    method: str

    #: Query arguments added to the URL.
    #:
    #: Type:
    #:     dict
    query: QueryArgs | None

    #: The URL the request is being made to.
    #:
    #: Type:
    #:     str
    url: str

    #: The list of handlers to use when making a HTTP(S) request.
    #:
    #: Type:
    #:     list
    _urlopen_handlers: list[BaseHandler]

    def __init__(
        self,
        url: str,
        query: (QueryArgs | None) = None,
        body: (bytes | None) = None,
        headers: (HTTPHeaders | None) = None,
        method: str = 'GET',
        hosting_service: (BaseHostingService | None) = None,
        **kwargs,
    ) -> None:
        """Initialize the request.

        Args:
            url (str):
                The URL to make the request against.

            query (dict, optional):
                Query arguments to add onto the URL.

                These will be mixed with any query arguments already in the
                URL, and the result will be applied in sorted order, for
                cross-Python compatibility.

            body (bytes, optional):
                The payload body for the request, if using a ``POST`` or
                ``PUT`` request.

            headers (dict, optional):
                Additional headers to attach to the request.

            method (str, optional):
                The request method. If not provided, it defaults to a ``GET``
                request.

            hosting_service (reviewboard.hostingsvcs.base.hosting_service.
                             BaseHostingService, optional):
                The hosting service this request is associated with.

            **kwargs (dict, unused):
                Additional keyword arguments for the request. This is unused,
                but allows room for expansion by subclasses.
        """
        # These must be set before any _log_and_raise() calls.
        self.method = method
        self.hosting_service = hosting_service

        if body is not None and not isinstance(body, bytes):
            _log_and_raise(
                body,
                self,
                'Received non-bytes body for the HTTP request for '
                '%(service)r. This is likely an implementation problem. '
                'Please make sure only byte strings are sent for the request '
                'body.')

        self.headers = {}

        if headers:
            for key, value in headers.items():
                self.add_header(key, value)

        if query:
            parsed_url = list(urlparse(url))
            new_query = parse_qs(parsed_url[4])
            new_query.update(query)

            parsed_url[4] = urlencode(
                OrderedDict(
                    pair
                    for pair in sorted(new_query.items(),
                                       key=lambda pair: pair[0])
                ),
                doseq=True)

            url = urlunparse(parsed_url)

        self.body = body
        self.url = url
        self.query = query

        self._urlopen_handlers = []

    @property
    def data(self) -> bytes | None:
        """The payload data for the request.

        Deprecated:
            4.0:
            This is deprecated in favor of the :py:attr:`body` attribute.
        """
        class_name = type(self).__name__

        RemovedInReviewBoard90Warning.warn(
            f'{class_name}.data is deprecated in favor of '
            f'{class_name}.body. This will be removed in Review Board 9.'
        )

        return self.body

    def add_header(
        self,
        name: str,
        value: str,
    ) -> None:
        """Add a header to the request.

        Args:
            name (str):
                The header name.

            value (str):
                The header value.
        """
        if not isinstance(name, str):
            _log_and_raise(
                name,
                self,
                'Received non-Unicode header name %(header)r '
                '(value=%(data)r) for the HTTP request for %(service)r. '
                'This is likely an implementation problem. Please make sure '
                'only Unicode strings are sent in request headers.',
                header=name,
                data=value)

        if not isinstance(value, str):
            _log_and_raise(
                value,
                self,
                'Received non-Unicode header value for %(header)r '
                '(value=%(data)r) for the HTTP request for %(service)r. This '
                'is likely an implementation problem. Please make sure only '
                'Unicode strings are sent in request headers.',
                header=name,
                data=value)

        self.headers[name.capitalize()] = value

    def get_header(
        self,
        name: str,
        default: (str | None) = None,
    ) -> str | None:
        """Return a header from the request.

        Args:
            name (str):
                The header name.

            default (str, optional):
                The default value if the header was not found.

        Returns:
            str:
            The header value.
        """
        assert isinstance(name, str), (
            f'{type(self).__name__}.get_header() requires a Unicode header '
            f'name')

        return self.headers.get(name.capitalize(), default)

    def add_basic_auth(
        self,
        username: bytes | str,
        password: bytes | str,
    ) -> None:
        """Add HTTP Basic Authentication headers to the request.

        Args:
            username (str or bytes):
                The username.

            password (str or bytes):
                The password.
        """
        if isinstance(username, str):
            username = username.encode('utf-8')

        if isinstance(password, str):
            password = password.encode('utf-8')

        auth = b'%s:%s' % (username, password)
        encoded = force_str(base64.b64encode(auth))

        self.add_header(HTTPBasicAuthHandler.auth_header,
                        f'Basic {encoded}')

    def add_digest_auth(
        self,
        username: str,
        password: str,
    ) -> None:
        """Add HTTP Digest Authentication support to the request.

        Args:
            username (str):
                The username.

            password (str):
                The password.
        """
        result = urlparse(self.url)
        top_level_url = f'{result.scheme}://{result.netloc}'

        password_mgr = HTTPPasswordMgrWithDefaultRealm()
        password_mgr.add_password(None, top_level_url, username, password)

        self.add_urlopen_handler(HTTPDigestAuthHandler(password_mgr))

    def add_urlopen_handler(
        self,
        handler: BaseHandler,
    ) -> None:
        """Add a handler to invoke for the urlopen call.

        Note:
            This is dependent on a :py:mod:`urllib`-backed request. While
            that is the default today, it may not be in the future. This
            method should be used with the knowledge that it may someday be
            deprecated, or may not work at all with special subclasses.

        Args:
            handler (urllib.request.AbstractHTTPHandler):
                The handler to add.
        """
        self._urlopen_handlers.append(handler)

    def open(self) -> HostingServiceHTTPResponse:
        """Open the request to the server, returning the response.

        Returns:
            HostingServiceHTTPResponse:
            The response information from the server.

        Raises:
            urllib.error.URLError:
                An error occurred talking to the server, or an HTTP error
                (400+) was returned.
        """
        url = self.url
        method = self.method

        request = BaseURLRequest(url=url,
                                 data=self.body,
                                 headers=self.headers,
                                 method=method)
        hosting_service = self.hosting_service
        urlopen_handlers = self._urlopen_handlers

        if hosting_service:
            hosting_account = hosting_service.account

            if 'ssl_cert' in hosting_account.data:
                # There's existing legacy SSL certificate data stored in
                # the account. Convert it to a modern Certificate if
                # possible, and build a context with it. If successful,
                # 'ssl_cert' will be removed from the data.
                ssl_kwargs = _build_ssl_kwargs_from_ssl_cert(
                    hosting_account=hosting_account,
                )
            else:
                ssl_kwargs = {}

            urlopen_handlers.append(CertificateVerificationHTTPSHandler(
                local_site=hosting_account.local_site,
                **ssl_kwargs,
            ))

            timer_msg = (
                f'Performing HTTP {method} request for '
                f'{hosting_service.name} at {url}'
            )
        else:
            timer_msg = (
                f'Performing HTTP {method} request at {url}'
            )

        opener = build_opener(*urlopen_handlers)

        with log_timed(timer_msg,
                       logger=logger):
            response = opener.open(request)

        if hosting_service:
            response_cls = hosting_service.client.http_response_cls
        else:
            response_cls = HostingServiceHTTPResponse

        return response_cls(request=self,
                            url=response.geturl(),
                            data=response.read(),
                            headers=dict(response.headers),
                            status_code=response.getcode())


class HostingServiceHTTPResponse:
    """An HTTP response from the server.

    This stores the URL, payload data, headers, and status code from an
    HTTP response.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.service` to
          :py:mod:`reviewboard.hostingsvcs.base.http`.
        * Deprecated legacy tuple representations of this class now emit
          deprecation warnings and will be removed in Review Board 8.

    Version Added:
        4.0
    """

    ######################
    # Instance variables #
    ######################

    #: The response data.
    #:
    #: Type:
    #:     bytes
    data: bytes

    #: The HTTP response headers.
    #:
    #: It's recommended to call :py:meth:`get_header` to request a header.
    #:
    #: Type:
    #:     dict
    headers: HTTPHeaders

    #: The HTTP request this is in response to.
    #:
    #: Type:
    #:     HostingServiceHTTPRequest
    request: HostingServiceHTTPRequest

    #: The HTTP status code for the response.
    #:
    #: Type:
    #:     int
    status_code: int

    #: The URL providing the response.
    url: str

    def __init__(
        self,
        request: HostingServiceHTTPRequest,
        url: str,
        data: bytes | None,
        headers: HTTPHeaders,
        status_code: int,
    ) -> None:
        """Initialize the response.

        Args:
            request (HostingServiceHTTPRequest):
                The request this is in response to.

            url (str):
                The URL serving the response.

                If redirected, this may differ from the request URL.

            data (bytes):
                The response payload.

            headers (dict):
                The response headers.

            status_code (int):
                The response HTTP status code.
        """
        self.request = request

        if data is None:
            data = b''
        elif not isinstance(data, bytes):
            # HTTP response data will be in byte strings, unless something is
            # overridden. Users should never see this in production, but
            # it'll be confusing for development. Make sure developers see
            # this through both a log message and an exception.
            _log_and_raise(
                data,
                request,
                'Received non-byte data from the HTTP %(method)s request '
                'for %(service)r. This is likely an implementation '
                'problem in a unit test or subclass. Please make sure '
                'only byte strings are sent.')

        if not isinstance(headers, dict):
            _log_and_raise(
                headers,
                request,
                'Headers response for HTTP %(method)s request for %(service)r '
                'is not a dict. This is likely an implementation problem in a '
                'unit test. Please make sure a dictionary is returned.')

        new_headers: HTTPHeaders = {}

        for key, value in headers.items():
            if not isinstance(key, str):
                _log_and_raise(
                    key,
                    request,
                    'Received non string header %(header)r from the '
                    'HTTP %(method)s request for %(service)r. This is likely '
                    'an implementation problem in a unit test. Please '
                    'make sure only strings are sent.',
                    header=key)

            if not isinstance(value, str):
                _log_and_raise(
                    value,
                    request,
                    'Received non string header value %(header)r '
                    '(key=%(key)r) from the HTTP %(method)s request for '
                    '%(service)r. This is likely an implementation problem in '
                    'a unit test. Please make sure only strings are  sent.',
                    key=key,
                    header=value)

            new_headers[key.capitalize()] = value

        self.url = url
        self.data = data
        self.headers = new_headers
        self.status_code = status_code

    @cached_property
    def json(self) -> JSONValue:
        """A JSON representation of the payload data.

        Raises:
            ValueError:
                The data is not valid JSON.
        """
        data = self.data

        if data:
            # There's actual data here, so parse it and return it.
            return json.loads(data.decode('utf-8'))

        # Return whatever falsey value we received.
        return data

    def get_header(
        self,
        name: str,
        default: (str | None) = None,
    ) -> str | None:
        """Return the value of a header as a Unicode string.

        This accepts a header name with any form of capitalization. The header
        name will be normalized.

        Args:
            name (str):
                The header name.

            default (str, optional):
                The default value if the header is not set.

        Returns:
            str:
            The resulting header value.
        """
        return self.headers.get(name.capitalize(), default)

    def __getitem__(
        self,
        i: int,
    ) -> Any:
        """Return an indexed item from the response.

        This is used to emulate the older 2-tuple response returned by hosting
        service HTTP request methods.

        Deprecated:
            6.0:
            Callers should instead access :py:attr:`data` or :py:attr:`headers`
            on this object.

        Args:
            i (int):
                The index of the item.

        Returns:
            object:
            The object at the specified index.

            If 0, this will return :py:attr:`data`.

            If 1, this will return :py:attr:`headers`.

        Raises:
            IndexError:
                An index other than 0 or 1 was requested.
        """
        class_name = type(self).__name__

        RemovedInReviewBoard90Warning.warn(
            f'Accessing {class_name} by index is deprecated. Please use '
            f'{class_name}.data or {class_name}.headers instead. This '
            'will be removed in Review Board 9.'
        )

        if i == 0:
            return self.data
        elif i == 1:
            return self.headers
        else:
            raise IndexError



"""Unit tests for reviewboard.datagrids.columns.UsernameColumn.

Version Added:
    5.0.7
"""

from __future__ import annotations

from django.contrib.auth.models import User
from django.utils.safestring import SafeString
from djblets.testing.decorators import add_fixtures

from reviewboard.datagrids.columns import UsernameColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class UsernameColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.UsernameColumn."""

    column = UsernameColumn()

    @add_fixtures(['test_site'])
    def test_render(self):
        """Testing UsernameColumn.render_cell"""
        user = User.objects.get(username='doc')
        self.assertIn(
            'href="/users/doc/"',
            self.column.render_cell(self.stateful_column, user, None))

    @add_fixtures(['test_site'])
    def test_render_local_site(self):
        """Testing UsernameColumn.render_cell on a LocalSite"""
        self.request._local_site_name = self.local_site_name
        user = User.objects.get(username='doc')

        self.assertIn(
            'href="/s/%s/users/doc/"' % self.local_site_name,
            self.column.render_cell(self.stateful_column, user, None))

    def test_render_data(self) -> None:
        """Testing UsernameColumn.render_data"""
        user = User.objects.get(username='doc')

        # The output includes the username along with avatar HTML. We verify
        # the username appears correctly; the full avatar HTML is dynamic.
        value = self.column.render_data(self.stateful_column, user)

        self.assertIsInstance(value, SafeString)
        self.assertIn('doc', value)

    def test_to_json(self) -> None:
        """Testing UsernameColumn.to_json"""
        user = User.objects.get(username='doc')

        # UsernameColumn uses db_field (not field_name), so the default
        # Column.to_json() via get_raw_object_value() returns None.
        self.assertIsNone(self.column.to_json(self.stateful_column, user))



"""Models for repository support."""

from __future__ import annotations

import logging
import uuid
from importlib import import_module
from time import time
from typing import (Any, ClassVar, Final, Mapping, Optional, Sequence,
                    TYPE_CHECKING, Union, cast)
from urllib.parse import quote

from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import IntegrityError, models
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy as _
from djblets.cache.backend import cache_memoize, make_cache_key
from djblets.db.fields import JSONField
from djblets.log import log_timed
from djblets.util.decorators import cached_property
from housekeeping import deprecate_non_keyword_only_args

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base import hosting_service_registry
from reviewboard.hostingsvcs.errors import MissingHostingServiceError
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.scmtools import scmtools_registry
from reviewboard.scmtools.core import FileLookupContext
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.managers import RepositoryManager, ToolManager
from reviewboard.scmtools.signals import (checked_file_exists,
                                          checking_file_exists,
                                          fetched_file, fetching_file)
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django.contrib.auth.models import AnonymousUser
    from django.http import HttpRequest
    from typelets.funcs import KwargsDict
    from typelets.django.strings import StrPromise

    from reviewboard.scmtools.core import Branch, Commit, SCMTool
    from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService


logger = logging.getLogger(__name__)


class Tool(models.Model):
    """A configured source code management tool.

    Each :py:class:`~reviewboard.scmtools.core.SCMTool` used by repositories
    must have a corresponding :py:class:`Tool` entry. These provide information
    on the capabilities of the tool, and accessors to construct a tool for
    a repository.

    Deprecated:
        5.0:
        This model is now obsolete. Any usage of this should be updated to use
        equivalent methods on the Repository or SCMTool instead.
    """

    name = models.CharField(max_length=32, unique=True)
    class_name = models.CharField(max_length=128, unique=True)

    objects: ClassVar[ToolManager] = ToolManager()

    # Templates can't access variables on a class properly. It'll attempt to
    # instantiate the class, which will fail without the necessary parameters.
    # So, we use these as convenient wrappers to do what the template can't do.

    #: Overridden help text for the configuration form fields.
    #:
    #: See :py:attr:`SCMTool.field_help_text
    #: <reviewboard.scmtools.core.SCMTool.field_help_text>` for details.
    field_help_text = property(
        lambda x: x.scmtool_class.field_help_text)

    @property
    def scmtool_id(self) -> str:
        """The unique ID for the SCMTool.

        Type:
            str
        """
        return self.scmtool_class.scmtool_id

    def get_scmtool_class(self) -> type[SCMTool]:
        """Return the configured SCMTool class.

        Returns:
            type:
            The subclass of :py:class:`~reviewboard.scmtools.core.SCMTool`
            backed by this Tool entry.

        Raises:
            django.core.exceptions.ImproperlyConfigured:
                The SCMTool could not be found.
        """
        if not hasattr(self, '_scmtool_class'):
            path = self.class_name
            i = path.rfind('.')
            module, attr = path[:i], path[i + 1:]

            try:
                mod = import_module(module)
            except ImportError as e:
                raise ImproperlyConfigured(
                    'Error importing SCM Tool %s: "%s"' % (module, e))

            try:
                self._scmtool_class = getattr(mod, attr)
            except AttributeError:
                raise ImproperlyConfigured(
                    'Module "%s" does not define a "%s" SCM Tool'
                    % (module, attr))

        return self._scmtool_class
    scmtool_class = property(get_scmtool_class)

    def __str__(self) -> str:
        """Return the name of the tool.

        Returns:
            str:
            The name of the tool.
        """
        return self.name

    class Meta:
        db_table = 'scmtools_tool'
        ordering = ('name',)
        verbose_name = _('Tool')
        verbose_name_plural = _('Tools')


class Repository(models.Model):
    """A configured external source code repository.

    Each configured Repository entry represents a source code repository that
    Review Board can communicate with as part of the diff uploading and
    viewing process.

    Repositories are backed by a
    :py:class:`~reviewboard.scmtools.core.SCMTool`, which functions as a client
    for the type of repository and can fetch files, load lists of commits and
    branches, and more.

    Access control is managed by a combination of the :py:attr:`public`,
    :py:attr:`users`, and :py:attr:`groups` fields. :py:attr:`public` controls
    whether a repository is publicly-accessible by all users on the server.
    When ``False``, only users explicitly listed in :py:attr:`users` or users
    who are members of the groups listed in :py:attr:`groups` will be able
    to access the repository or view review requests posted against it.
    """

    #: The amount of time branches are cached, in seconds.
    #:
    #: Branches are cached for 5 minutes.
    BRANCHES_CACHE_PERIOD: Final[int] = 60 * 5

    #: The short period of time to cache commit information, in seconds.
    #:
    #: Some commit information (such as retrieving the latest commits in a
    #: repository) should result in information cached only for a short
    #: period of time. This is set to cache for 5 minutes.
    COMMITS_CACHE_PERIOD_SHORT: Final[int] = 60 * 5

    #: The long period of time to cache commit information, in seconds.
    #:
    #: Commit information that is unlikely to change should be kept around
    #: for a longer period of time. This is set to cache for 1 day.
    COMMITS_CACHE_PERIOD_LONG: Final[int] = 60 * 60 * 24  # 1 day

    #: The fallback encoding for text-based files in repositories.
    #:
    #: This is used if the file isn't valid UTF-8, and if the repository
    #: doesn't specify a list of encodings.
    FALLBACK_ENCODING: Final[str] = 'iso-8859-15'

    #: The error message used to indicate that a repository name conflicts.
    NAME_CONFLICT_ERROR: Final[StrPromise] = \
        _('A repository with this name already exists')

    #: The error message used to indicate that a repository path conflicts.
    PATH_CONFLICT_ERROR: Final[StrPromise] = \
        _('A repository with this path already exists')

    #: The prefix used to indicate an encrypted password.
    #:
    #: This is used to indicate whether a stored password is in encrypted
    #: form or plain text form.
    ENCRYPTED_PASSWORD_PREFIX: Final[str] = '\t'

    name = models.CharField(_('Name'), max_length=255)
    path = models.CharField(_('Path'), max_length=255)
    mirror_path = models.CharField(max_length=255, blank=True)
    raw_file_url = models.CharField(
        _('Raw file URL mask'),
        max_length=255,
        blank=True)
    username = models.CharField(max_length=32, blank=True)
    encrypted_password = models.CharField(max_length=128, blank=True,
                                          db_column='password')
    extra_data = JSONField(null=True)

    tool = models.ForeignKey(Tool, on_delete=models.CASCADE,
                             related_name='repositories')

    scmtool_id = models.CharField(max_length=255, null=True, blank=True)

    hosting_account = models.ForeignKey(
        HostingServiceAccount,
        on_delete=models.CASCADE,
        related_name='repositories',
        verbose_name=_('Hosting service account'),
        blank=True,
        null=True)

    bug_tracker = models.CharField(
        _('Bug tracker URL'),
        max_length=256,
        blank=True,
        help_text=_("This should be the full path to a bug in the bug tracker "
                    "for this repository, using '%s' in place of the bug ID."))
    encoding = models.CharField(
        max_length=32,
        blank=True,
        help_text=_("The encoding used for files in this repository. This is "
                    "an advanced setting and should only be used if you're "
                    "sure you need it."))
    visible = models.BooleanField(
        _('Show this repository'),
        default=True,
        help_text=_('Use this to control whether or not a repository is '
                    'shown when creating new review requests. Existing '
                    'review requests are unaffected.'))

    archived = models.BooleanField(
        _('Archived'),
        default=False,
        help_text=_("Archived repositories do not show up in lists of "
                    "repositories, and aren't open to new review requests."))

    archived_timestamp = models.DateTimeField(null=True, blank=True)

    # Access control
    local_site = models.ForeignKey(LocalSite,
                                   on_delete=models.CASCADE,
                                   verbose_name=_('Local site'),
                                   blank=True,
                                   null=True)
    public = models.BooleanField(
        _('publicly accessible'),
        default=True,
        help_text=_('Review requests and files on public repositories are '
                    'visible to anyone. Private repositories must explicitly '
                    'list the users and groups that can access them.'))

    users = models.ManyToManyField(
        User,
        limit_choices_to={'is_active': True},
        blank=True,
        related_name='repositories',
        verbose_name=_('Users with access'),
        help_text=_('A list of users with explicit access to the repository.'))
    review_groups = models.ManyToManyField(
        'reviews.Group',
        limit_choices_to={'invite_only': True},
        blank=True,
        related_name='repositories',
        verbose_name=_('Review groups with access'),
        help_text=_('A list of invite-only review groups whose members have '
                    'explicit access to the repository.'))

    hooks_uuid = models.CharField(
        _('Hooks UUID'),
        max_length=32,
        null=True,
        blank=True,
        help_text=_('Unique identifier used for validating incoming '
                    'webhooks.'))

    objects: ClassVar[RepositoryManager] = RepositoryManager()

    ######################
    # Instance variables #
    ######################

    #: The cached SCMTool class backing this repository.
    #:
    #: This is only present once :py:attr:`scmtool_class` has been invoked
    #: successfully.
    _scmtool_class: type[SCMTool]

    @property
    def password(self) -> Optional[str]:
        """The password for the repository.

        If a password is stored and encrypted, it will be decrypted and
        returned.

        If the stored password is in plain-text, then it will be encrypted,
        stored in the database, and returned.
        """
        password: Optional[str] = self.encrypted_password

        # NOTE: Due to a bug in 2.0.9, it was possible to get a string of
        #       "\tNone", indicating no password. We have to check for this.
        if not password or password == '\tNone':
            password = None
        elif password.startswith(self.ENCRYPTED_PASSWORD_PREFIX):
            password = password[len(self.ENCRYPTED_PASSWORD_PREFIX):]

            if password:
                try:
                    password = decrypt_password(password)
                except ValueError:
                    logger.critical('Unable to decrypt stored password for '
                                    'repository pk=%d', self.pk)
                    password = None
            else:
                password = None
        else:
            # This is a plain-text password. Convert it.
            self.password = password
            self.save(update_fields=['encrypted_password'])

        return password

    @password.setter
    def password(
        self,
        value: Optional[str],
    ) -> None:
        """Set the password for the repository.

        The password will be stored as an encrypted value, prefixed with a
        tab character in order to differentiate between legacy plain-text
        passwords.

        Args:
            password (str):
                The new password to set.
        """
        if value:
            value = '%s%s' % (self.ENCRYPTED_PASSWORD_PREFIX,
                              encrypt_password(value))
        else:
            value = ''

        self.encrypted_password = value

    @property
    def scmtool_class(self) -> Optional[type[SCMTool]]:
        """The SCMTool subclass used for this repository.

        Type:
            type:
            A subclass of :py:class:`~reviewboard.scmtools.core.SCMTool`.

        Raises:
            django.core.exceptions.ImproperlyConfigured:
                The SCMTool could not be found, due to missing packages or
                extensions. Details are in the message, and the failure is
                logged.
        """
        # We'll optimistically cache this, mirroring behavior in
        # Tool.get_scmtool_class(). Note that we only cache below if we get
        # a non-None value, as None can occur while an instance is being set
        # up.
        if hasattr(self, '_scmtool_class'):
            return self._scmtool_class

        scmtool_id = self.scmtool_id

        if scmtool_id:
            tool = scmtools_registry.get_by_id(scmtool_id)

            if tool is not None:
                self._scmtool_class = tool
                return tool

            logger.error('Error finding registered SCMTool "%s" in '
                         'repository ID %s.',
                         scmtool_id, self.pk)
        elif not self.tool_id:
            # For backwards-compatibility reasons, we return None when there's
            # no Tool object associated.
            return None

        # We use ImproperlyConfigured here for compatibility with the
        # the call to Tool.get_scmtool_class() in Review Board < 5.0.
        raise ImproperlyConfigured(
            gettext(
                'There was an error loading the SCMTool "%s" needed by this '
                'repository. The administrator should ensure all necessary '
                'packages and extensions are installed.'
            )
            % (scmtool_id or self.tool.name))

    @cached_property
    def hosting_service(self) -> Optional[BaseHostingService]:
        """The hosting service providing this repository.

        This will be ``None`` if this is a standalone repository.

        Type:
            reviewboard.hostingsvcs.base.hosting_service.BaseHostingService

        Raises:
            reviewboard.hostingsvcs.errors.MissingHostingServiceError:
                The hosting service for this repository could not be loaded.
        """
        if self.hosting_account:
            try:
                return self.hosting_account.service
            except MissingHostingServiceError as e:
                raise MissingHostingServiceError(e.hosting_service_id,
                                                 self.name)

        return None

    @cached_property
    def bug_tracker_service(self) -> Optional[BaseHostingService]:
        """The selected bug tracker service for the repository.

        This will be ``None`` if this repository is not associated with a bug
        tracker.

        Type:
            reviewboard.hostingsvcs.base.hosting_service.BaseHostingService

        Raises:
            reviewboard.hostingsvcs.errors.MissingHostingServiceError:
                The hosting service for this repository could not be loaded.
        """
        if self.extra_data.get('bug_tracker_use_hosting'):
            return self.hosting_service

        bug_tracker_type = self.extra_data.get('bug_tracker_type')

        if bug_tracker_type:
            bug_tracker_cls = \
                hosting_service_registry.get_hosting_service(bug_tracker_type)
            assert bug_tracker_cls is not None

            # TODO: we need to figure out some way of storing a second
            # hosting service account for bug trackers.
            return bug_tracker_cls(HostingServiceAccount())

        return None

    @property
    def supports_post_commit(self) -> bool:
        """Whether or not this repository supports post-commit creation.

        If this is ``True``, the :py:meth:`get_branches` and
        :py:meth:`get_commits` methods will be implemented to fetch information
        about the committed revisions, and get_change will be implemented to
        fetch the actual diff. This is used by
        :py:meth:`ReviewRequestDraft.update_from_commit_id
        <reviewboard.reviews.models.ReviewRequestDraft.update_from_commit_id>`.

        Type:
            bool

        Raises:
            reviewboard.hostingsvcs.errors.MissingHostingServiceError:
                The hosting service for this repository could not be loaded.
        """
        hosting_service = self.hosting_service

        if hosting_service:
            return hosting_service.supports_post_commit
        else:
            scmtool_class = self.scmtool_class
            assert scmtool_class is not None

            return scmtool_class.supports_post_commit

    @property
    def supports_pending_changesets(self) -> bool:
        """Whether this repository supports server-aware pending changesets.

        Type:
            bool
        """
        scmtool_class = self.scmtool_class
        assert scmtool_class is not None

        return scmtool_class.supports_pending_changesets

    @cached_property
    def diffs_use_absolute_paths(self) -> bool:
        """Whether or not diffs for this repository contain absolute paths.

        Some types of source code management systems generate diffs that
        contain paths relative to the directory where the diff was generated.
        Most contain absolute paths. This flag indicates which path format
        this repository can expect.

        Type:
            bool
        """
        # Ideally, we won't have to instantiate the class, as that can end up
        # performing some expensive calls or HTTP requests.  If the SCMTool is
        # modern (doesn't define a get_diffs_use_absolute_paths), it will have
        # all the information we need on the class. If not, we might have to
        # instantiate it, but do this as a last resort.
        scmtool_cls = self.scmtool_class
        assert scmtool_cls is not None

        if isinstance(scmtool_cls.diffs_use_absolute_paths, bool):
            return scmtool_cls.diffs_use_absolute_paths
        elif hasattr(scmtool_cls, 'get_diffs_use_absolute_paths'):
            # This will trigger a deprecation warning.
            return self.get_scmtool().diffs_use_absolute_paths
        else:
            return False

    def get_scmtool(self) -> SCMTool:
        """Return an instance of the SCMTool for this repository.

        Each call will construct a brand new instance. The returned value
        should be stored and used for multiple operations in a single session.

        Returns:
            reviewboard.scmtools.core.SCMTool:
            A new instance of the SCMTool for this repository.
        """
        scmtool_class = self.scmtool_class
        assert scmtool_class is not None

        return scmtool_class(self)

    def get_credentials(self) -> Mapping[str, Any]:
        """Return the credentials for this repository.

        This returns a dictionary with ``username`` and ``password`` keys.
        By default, these will be the values stored for the repository,
        but if a hosting service is used and the repository doesn't have
        values for one or both of these, the hosting service's credentials
        (if available) will be used instead.

        Returns:
            dict:
            A dictionary with credentials information.
        """
        username = self.username
        password = self.password
        hosting_account = self.hosting_account

        if hosting_account and hosting_account.service:
            username = username or hosting_account.username
            password = password or hosting_account.service.get_password()

        return {
            'username': username,
            'password': password,
        }

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def get_or_create_hooks_uuid(
        self,
        *,
        max_attempts: int = 20,
    ) -> str:
        """Return a hooks UUID, creating one if necessary.

        A hooks UUID is used for creating unique incoming webhook URLs,
        allowing services to communicate information to Review Board.

        If a hooks UUID isn't already saved, then this will try to generate one
        that doesn't conflict with any other registered hooks UUID. It will try
        up to ``max_attempts`` times, and if it fails, ``None`` will be
        returned.

        Args:
            max_attempts (int, optional):
                The maximum number of UUID generation attempts to try before
                giving up.

        Returns:
            str:
            The resulting UUID.

        Raises:
            Exception:
                The maximum number of attempts has been reached.
        """
        if not self.hooks_uuid:
            for attempt in range(max_attempts):
                self.hooks_uuid = uuid.uuid4().hex

                try:
                    self.save(update_fields=['hooks_uuid'])
                    break
                except IntegrityError:
                    # We hit a collision with the token value. Try again.
                    self.hooks_uuid = None

            if not self.hooks_uuid:
                s = ('Unable to generate a unique hooks UUID for '
                     'repository %s after %d attempts'
                     % (self.pk, max_attempts))
                logger.error(s)
                raise Exception(s)

        return self.hooks_uuid

    def get_encoding_list(self) -> Sequence[str]:
        """Return a list of candidate text encodings for files.

        This will return a list based on a comma-separated list of encodings
        in :py:attr:`encoding`. If no encodings are configured, the default
        of ``iso-8859-15`` will be used.

        Returns:
            list of str:
            The list of text encodings to try for files in the repository.
        """
        encodings: list[str] = []

        for e in self.encoding.split(','):
            e = e.strip()

            if e:
                encodings.append(e)

        return encodings or [self.FALLBACK_ENCODING]

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def get_file(
        self,
        *,
        path: str,
        revision: str,
        base_commit_id: Optional[str] = None,
        request: Optional[HttpRequest] = None,
        context: Optional[FileLookupContext] = None,
    ) -> bytes:
        """Return a file from the repository.

        This will attempt to retrieve the file from the repository. If the
        repository is backed by a hosting service, it will go through that.
        Otherwise, it will attempt to directly access the repository.

        This will send the
        :py:data:`~reviewboard.scmtools.signals.fetching_file` signal before
        beginning a file fetch from the repository (if not cached), and the
        :py:data:`~reviewboard.scmtools.signals.fetched_file` signal after.

        Args:
            path (str):
                The path to the file in the repository.

            revision (str):
                The revision of the file to retrieve.

            base_commit_id (str, optional):
                The ID of the commit containing the revision of the file
                to retrieve. This is required for some types of repositories
                where the revision of a file and the ID of a commit differ.

                Deprecated:
                    4.0.5:
                    Callers should provide this in ``context`` instead.

            request (django.http.HttpRequest, optional):
                The current HTTP request from the client. This is used for
                logging purposes.

                Deprecated:
                    4.0.5:
                    Callers should provide this in ``context`` instead.

            context (reviewboard.scmtools.core.FileLookupContext, optional):
                Extra context used to help look up this file.

                This contains information about the HTTP request, requesting
                user, and parsed diff information, which may be useful as
                part of the repository lookup process.

                Version Added:
                    4.0.5

        Returns:
            bytes:
            The resulting file contents.

        Raises:
            TypeError:
                One or more of the provided arguments is an invalid type.
                Details are contained in the error message.
        """
        # We wrap the result of get_file in a list and then return the first
        # element after getting the result from the cache. This prevents the
        # cache backend from converting to unicode, since we're no longer
        # passing in a string and the cache backend doesn't recursively look
        # through the list in order to convert the elements inside.
        #
        # Basically, this fixes the massive regressions introduced by the
        # Django unicode changes.
        if not isinstance(path, str):
            raise TypeError('"path" must be a Unicode string, not %s'
                            % type(path))

        if not isinstance(revision, str):
            raise TypeError('"revision" must be a Unicode string, not %s'
                            % type(revision))

        if context is None:
            # If an explicit context isn't provided, create one. In a future
            # version, this will be required.
            context = FileLookupContext(request=request,
                                        base_commit_id=base_commit_id)

        return cache_memoize(
            self._make_file_cache_key(path=path,
                                      revision=revision,
                                      base_commit_id=context.base_commit_id),
            lambda: [
                self._get_file_uncached(path=path,
                                        revision=revision,
                                        context=context),
            ],
            large_data=True)[0]

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def get_file_exists(
        self,
        *,
        path: str,
        revision: str,
        base_commit_id: Optional[str] = None,
        request: Optional[HttpRequest] = None,
        context: Optional[FileLookupContext] = None,
    ) -> bool:
        """Return whether or not a file exists in the repository.

        If the repository is backed by a hosting service, this will go
        through that. Otherwise, it will attempt to directly access the
        repository.

        The result of this call will be cached, making future lookups
        of this path and revision on this repository faster.

        This will send the
        :py:data:`~reviewboard.scmtools.signals.checking_file_exists` signal
        before beginning a file fetch from the repository (if not cached), and
        the :py:data:`~reviewboard.scmtools.signals.checked_file_exists` signal
        after.

        Args:
            path (str):
                The path to the file in the repository.

            revision (str);
                The revision of the file to check.

            base_commit_id (str, optional):
                The ID of the commit containing the revision of the file
                to check. This is required for some types of repositories
                where the revision of a file and the ID of a commit differ.

                Deprecated:
                    4.0.5:
                    Callers should provide this in ``context`` instead.

            request (django.http.HttpRequest, optional):
                The current HTTP request from the client. This is used for
                logging purposes.

                Deprecated:
                    4.0.5:
                    Callers should provide this in ``context`` instead.

            context (reviewboard.scmtools.core.FileLookupContext, optional):
                Extra context used to help look up this file.

                This contains information about the HTTP request, requesting
                user, and parsed diff information, which may be useful as
                part of the repository lookup process.

                Version Added:
                    4.0.5

        Returns:
            bool:
            ``True`` if the file exists in the repository. ``False`` if it
            does not.

        Raises:
            TypeError:
                One or more of the provided arguments is an invalid type.
                Details are contained in the error message.
        """
        if not isinstance(path, str):
            raise TypeError('"path" must be a Unicode string, not %s'
                            % type(path))

        if not isinstance(revision, str):
            raise TypeError('"revision" must be a Unicode string, not %s'
                            % type(revision))

        if context is None:
            # If an explicit context isn't provided, create one. In a future
            # version, this will be required.
            context = FileLookupContext(request=request,
                                        base_commit_id=base_commit_id)

        key = self._make_file_exists_cache_key(
            path=path,
            revision=revision,
            base_commit_id=context.base_commit_id)

        if cache.get(make_cache_key(key)) == '1':
            return True

        exists = self._get_file_exists_uncached(path=path,
                                                revision=revision,
                                                context=context)

        if exists:
            cache_memoize(key, lambda: '1')

        return exists

    def get_branches(self) -> Sequence[Branch]:
        """Return a list of all branches on the repository.

        This will fetch a list of all known branches for use in the API and
        New Review Request page.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The list of branches in the repository. One (and only one) will
            be marked as the default branch.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                The hosting service backing the repository encountered an
                error.

            reviewboard.scmtools.errors.SCMError:
                The repository tool encountered an error.

            NotImplementedError:
                Branch retrieval is not available for this type of repository.
        """
        def _get_branches() -> Sequence[Branch]:
            hosting_service = self.hosting_service

            with log_timed(f'Fetching branches from {self}',
                           logger=logger):
                if hosting_service:
                    return hosting_service.get_branches(self)
                else:
                    return self.get_scmtool().get_branches()

        cache_key = make_cache_key(f'repository-branches:{self.pk}')

        return cache_memoize(cache_key, _get_branches,
                             expiration=self.BRANCHES_CACHE_PERIOD)

    def get_commit_cache_key(
        self,
        commit_id: str,
    ) -> str:
        """Return the cache key used for a commit ID.

        The resulting cache key is used to cache information about a commit
        retrieved from the repository that matches the provided ID. This can
        be used to delete information already in cache.

        Args:
            commit_id (str):
                The ID of the commit to generate a cache key for.

        Returns:
            str:
            The resulting cache key.
        """
        return f'repository-commit:{self.pk}:{commit_id}'

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def get_commits(
        self,
        *,
        branch: Optional[str] = None,
        start: Optional[str] = None,
    ) -> Sequence[Commit]:
        """Return a list of commits.

        This will fetch a batch of commits from the repository for use in the
        API and New Review Request page.

        The resulting commits will be in order from newest to oldest, and
        should return upwards of a fixed number of commits (usually 30, but
        this depends on the type of repository and its limitations). It may
        also be limited to commits that exist on a given branch (if supported
        by the repository).

        This can be called multiple times in succession using the
        :py:attr:`Commit.parent` of the last entry as the ``start`` parameter
        in order to paginate through the history of commits in the repository.

        Args:
            branch (str, optional):
                The branch to limit commits to. This may not be supported by
                all repositories.

            start (str, optional):
                The commit to start at. If not provided, this will fetch the
                first commit in the repository.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The retrieved commits.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                The hosting service backing the repository encountered an
                error.

            reviewboard.scmtools.errors.SCMError:
                The repository tool encountered an error.

            NotImplementedError:
                Commits retrieval is not available for this type of repository.
        """
        def _get_commits() -> Sequence[Commit]:
            hosting_service = self.hosting_service

            commits_kwargs: KwargsDict = {
                'branch': branch,
                'start': start,
            }

            with log_timed(f'Fetching commits (branch={branch}, '
                           f'start={start} from {self}',
                           logger=logger):
                if hosting_service:
                    return hosting_service.get_commits(self, **commits_kwargs)
                else:
                    return self.get_scmtool().get_commits(**commits_kwargs)

        # We cache both the entire list for 'start', as well as each individual
        # commit. This allows us to reduce API load when people are looking at
        # the "new review request" page more frequently than they're pushing
        # code, and will usually save 1 API request when they go to actually
        # create a new review request.
        if branch and start:
            cache_period = self.COMMITS_CACHE_PERIOD_LONG
        else:
            cache_period = self.COMMITS_CACHE_PERIOD_SHORT

        cache_key = make_cache_key(
            f'repository-commits:{self.pk}:{branch}:{start}')
        commits = cache_memoize(cache_key, _get_commits,
                                expiration=cache_period)

        for commit in commits:
            assert commit.id is not None

            cache.set(self.get_commit_cache_key(commit.id),
                      commit, self.COMMITS_CACHE_PERIOD_LONG)

        return commits

    def get_change(
        self,
        revision: str,
    ) -> Commit:
        """Return an individual change/commit in the repository.

        Args:
            revision (str):
                The commit ID or revision to retrieve.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit from the repository.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                The hosting service backing the repository encountered an
                error.

            reviewboard.scmtools.errors.SCMError:
                The repository tool encountered an error.

            NotImplementedError:
                Commits retrieval is not available for this type of repository.
        """
        hosting_service = self.hosting_service

        with log_timed(f'Fetching change {revision} from {self}',
                       logger=logger):
            if hosting_service:
                return hosting_service.get_change(self, revision)
            else:
                return self.get_scmtool().get_change(revision)

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def normalize_patch(
        self,
        patch: bytes,
        *,
        filename: str,
        revision: str,
    ) -> bytes:
        """Normalize a diff/patch file before it's applied.

        This can be used to take an uploaded diff file and modify it so that
        it can be properly applied. This may, for instance, uncollapse
        keywords or remove metadata that would confuse :command:`patch`.

        This passes the request on to the hosting service or repository
        tool backend.

        Args:
            patch (bytes):
                The diff/patch file to normalize.

            filename (str):
                The name of the file being changed in the diff.

            revision (str):
                The revision of the file being changed in the diff.

        Returns:
            bytes:
            The resulting diff/patch file.

        Raises:
            reviewboard.hostingsvcs.errors.MissingHostingServiceError:
                The hosting service for this repository could not be loaded.
        """
        hosting_service = self.hosting_service

        if hosting_service:
            return hosting_service.normalize_patch(repository=self,
                                                   patch=patch,
                                                   filename=filename,
                                                   revision=revision)
        else:
            return self.get_scmtool().normalize_patch(patch=patch,
                                                      filename=filename,
                                                      revision=revision)

    def is_accessible_by(
        self,
        user: Union[AnonymousUser, User],
    ) -> bool:
        """Return whether or not the user has access to the repository.

        The repository is accessibly by the user if it is public or
        the user has access to it (either by being explicitly on the allowed
        users list, or by being a member of a review group on that list).

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            ``True`` if the repository is accessible by the user.
            ``False`` if it is not.
        """
        if self.local_site and not self.local_site.is_accessible_by(user):
            return False

        if self.public or user.is_superuser:
            return True

        return (user.is_authenticated and
                (self.review_groups.filter(users__pk=user.pk).exists() or
                 self.users.filter(pk=user.pk).exists()))

    def is_mutable_by(
        self,
        user: Union[AnonymousUser, User],
    ) -> bool:
        """Return whether or not the user can modify or delete the repository.

        The repository is mutable by the user if the user is an administrator
        with proper permissions or the repository is part of a LocalSite and
        the user has permissions to modify it.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            ``True`` if the repository can modify or delete the repository.
            ``False`` if they cannot.
        """
        return user.has_perm('scmtools.change_repository', self.local_site)

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def archive(
        self,
        *,
        save: bool = True,
    ) -> None:
        """Archive a repository.

        The repository won't appear in any public lists of repositories,
        and won't be used when looking up repositories. Review requests
        can't be posted against an archived repository.

        New repositories can be created with the same information as the
        archived repository.

        Args:
            save (bool, optional):
                Whether to save the repository immediately.
        """
        # This should be sufficiently unlikely to create duplicates. time()
        # will use up a max of 8 characters, so we slice the name down to
        # make the result fit in 64 characters
        name_field = cast(models.CharField, self._meta.get_field('name'))
        max_name_len = name_field.max_length
        assert max_name_len is not None

        encoded_time = '%x' % int(time())
        reserved_len = len('ar::') + len(encoded_time)

        self.name = 'ar:%s:%s' % (self.name[:max_name_len - reserved_len],
                                  encoded_time)
        self.archived = True
        self.public = False
        self.archived_timestamp = timezone.now()

        if save:
            self.save(update_fields=('name', 'archived', 'public',
                                     'archived_timestamp'))

    def save(self, *args, **kwargs) -> None:
        """Save the repository.

        This will perform any data normalization needed, and then save the
        repository to the database.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the parent method.
        """
        # Prevent empty strings from saving in the admin UI, which could lead
        # to database-level validation errors.
        if self.hooks_uuid == '':
            self.hooks_uuid = None

        return super().save(*args, **kwargs)

    def clean(self) -> None:
        """Clean method for checking null unique_together constraints.

        Django has a bug where unique_together constraints for foreign keys
        aren't checked properly if one of the relations is null. This means
        that users who aren't using local sites could create multiple groups
        with the same name.

        Raises:
            django.core.exceptions.ValidationError:
                Validation of the model's data failed. Details are in the
                exception.
        """
        super().clean()

        if self.local_site is None:
            existing_repos = (
                Repository.objects
                .exclude(pk=self.pk)
                .filter(Q(name=self.name) |
                        (Q(archived=False) &
                         Q(path=self.path)))
                .values('name', 'path')
            )

            errors: dict[str, list[ValidationError]] = {}

            for repo_info in existing_repos:
                if repo_info['name'] == self.name:
                    errors['name'] = [
                        ValidationError(self.NAME_CONFLICT_ERROR,
                                        code='repository_name_exists'),
                    ]

                if repo_info['path'] == self.path:
                    errors['path'] = [
                        ValidationError(self.PATH_CONFLICT_ERROR,
                                        code='repository_path_exists'),
                    ]

            if errors:
                raise ValidationError(errors)

    def _make_file_cache_key(
        self,
        *,
        path: str,
        revision: str,
        base_commit_id: Optional[str],
    ) -> str:
        """Return a cache key for fetched files.

        Args:
            path (str):
                The path to the file in the repository.

            revision (str):
                The revision of the file.

            base_commit_id (str):
                The ID of the commit containing the revision of the file.
                This is required for some types of repositories where the
                revision of a file and the ID of a commit differ.

        Returns:
            str:
            A cache key representing this file.
        """
        return 'file:%s:%s:%s:%s:%s' % (
            self.pk,
            quote(path),
            quote(revision),
            quote(base_commit_id or ''),
            quote(self.raw_file_url or ''))

    def _make_file_exists_cache_key(
        self,
        *,
        path: str,
        revision: str,
        base_commit_id: Optional[str],
    ) -> str:
        """Makes a cache key for file existence checks.

        Args:
            path (str):
                The path to the file in the repository.

            revision (str);
                The revision of the file to check.

            base_commit_id (str, optional):
                The ID of the commit containing the revision of the file
                to check. This is required for some types of repositories
                where the revision of a file and the ID of a commit differ.

        Returns:
            str:
            A cache key representing this file check.
        """
        return 'file-exists:%s:%s:%s:%s:%s' % (
            self.pk,
            quote(path),
            quote(revision),
            quote(base_commit_id or ''),
            quote(self.raw_file_url or ''))

    def _get_file_uncached(
        self,
        *,
        path: str,
        revision: str,
        context: FileLookupContext
    ) -> bytes:
        """Return a file from the repository, bypassing cache.

        This is called internally by :py:meth:`get_file` if the file isn't
        already in the cache.

        This will send the
        :py:data:`~reviewboard.scmtools.signals.fetching_file` signal before
        beginning a file fetch from the repository, and the
        :py:data:`~reviewboard.scmtools.signals.fetched_file` signal after.

        Args:
            path (str):
                The path to the file in the repository.

            revision (str):
                The revision of the file to retrieve.

            context (reviewboard.scmtools.core.FileLookupContext):
                Extra context used to help look up this file.

                Version Added:
                    4.0.5

        Returns:
            bytes:
            The resulting file contents.

        Raises:
            reviewboard.hostingsvcs.errors.MissingHostingServiceError:
                The hosting service for this repository could not be loaded.
        """
        request = context.request
        base_commit_id = context.base_commit_id

        fetching_file.send(sender=self,
                           path=path,
                           revision=revision,
                           base_commit_id=base_commit_id,
                           request=request,
                           context=context)

        if base_commit_id:
            timer_msg = (
                f'Fetching file "{path}" r{revision} (base commit ID '
                f'{base_commit_id}) from {self}'
            )
        else:
            timer_msg = f'Fetching file "{path}" r{revision} from {self}'

        with log_timed(timer_msg,
                       logger=logger,
                       request=request):
            hosting_service = self.hosting_service

            if hosting_service:
                data = hosting_service.get_file(
                    self,
                    path,
                    revision,
                    base_commit_id=base_commit_id,
                    context=context)

                assert isinstance(data, bytes), (
                    '%s.get_file() must return a byte string, not %s'
                    % (type(hosting_service).__name__, type(data)))
            else:
                tool = self.get_scmtool()
                data = tool.get_file(path, revision,
                                     base_commit_id=base_commit_id,
                                     context=context)

                assert isinstance(data, bytes), (
                    '%s.get_file() must return a byte string, not %s'
                    % (type(tool).__name__, type(data)))

        fetched_file.send(sender=self,
                          path=path,
                          revision=revision,
                          base_commit_id=base_commit_id,
                          request=request,
                          context=context,
                          data=data)

        return data

    def _get_file_exists_uncached(
        self,
        *,
        path: str,
        revision: str,
        context: FileLookupContext,
    ) -> bool:
        """Check for file existence, bypassing cache.

        This is called internally by :py:meth:`get_file_exists` if the file
        isn't already in the cache.

        This function is smart enough to check if the file exists in cache,
        and will use that for the result instead of making a separate call.

        This will send the
        :py:data:`~reviewboard.scmtools.signals.checking_file_exists` signal
        before beginning a file fetch from the repository, and the
        :py:data:`~reviewboard.scmtools.signals.checked_file_exists` signal
        after.

        Args:
            path (str):
                The path to the file in the repository.

            revision (str):
                The revision of the file to check.

            context (reviewboard.scmtools.core.FileLookupContext):
                Extra context used to help look up this file.

                Version Added:
                    4.0.5

        Returns:
            bool:
            ``True`` if the file exists. ``False`` if it does not.

        Raises:
            reviewboard.hostingsvcs.errors.MissingHostingServiceError:
                The hosting service for this repository could not be loaded.
        """
        request = context.request
        base_commit_id = context.base_commit_id

        # First we check to see if we've fetched the file before. If so,
        # it's in there and we can just return that we have it.
        file_cache_key = make_cache_key(
            self._make_file_cache_key(path=path,
                                      revision=revision,
                                      base_commit_id=base_commit_id))

        if file_cache_key in cache:
            exists = True
        else:
            # We didn't have that in the cache, so check from the repository.
            checking_file_exists.send(sender=self,
                                      path=path,
                                      revision=revision,
                                      base_commit_id=base_commit_id,
                                      request=request,
                                      context=context)

            hosting_service = self.hosting_service

            if base_commit_id:
                timer_msg = (
                    f'Checking file existence for "{path}" r{revision} '
                    f'(base commit ID {base_commit_id}) from {self}'
                )
            else:
                timer_msg = (
                    f'Checking file existence for "{path}" r{revision} '
                    f'from {self}'
                )

            with log_timed(timer_msg,
                           logger=logger,
                           request=request):
                if hosting_service:
                    exists = hosting_service.get_file_exists(
                        self,
                        path,
                        revision,
                        base_commit_id=base_commit_id,
                        context=context)
                else:
                    tool = self.get_scmtool()
                    exists = tool.file_exists(path, revision,
                                              base_commit_id=base_commit_id,
                                              context=context)

            checked_file_exists.send(sender=self,
                                     path=path,
                                     revision=revision,
                                     base_commit_id=base_commit_id,
                                     request=request,
                                     exists=exists,
                                     context=context)

        return exists

    def __str__(self) -> str:
        """Return a string representation of the repository.

        This uses the repository's name as the string representation. However,
        it should not be used if explicitly wanting to retrieve the repository
        name, as future versions may return a different value.

        Returns:
            str:
            The repository name.
        """
        return self.name

    class Meta:
        db_table = 'scmtools_repository'
        unique_together = (('name', 'local_site'),
                           ('archived_timestamp', 'path', 'local_site'),
                           ('hooks_uuid', 'local_site'))
        verbose_name = _('Repository')
        verbose_name_plural = _('Repositories')
